
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <sys/types.h>
#include <iostream.h>
#include <fstream.h>
#include <assert.h>


#include "boot.h"
#include "statfcts.h"
#include "random.h"
#include "tisu.h"
#include "sort.h"
#include "math.h"
#include "vector.h"
#include "matrix.h"
#include "randlib.h"
#include "cdflib.h"
#include "kernel.h"


VectorDouble bootstrap(const VectorDouble &x, int nSame, int nStar)
// returns i.i.d. bootstrap pseudo sequence
{
  int n = x.getSize();
  if (nSame) nStar = n;
  VectorDouble xStar(nStar);

  for (int i = 0; i < nStar; i++)
    xStar[i] = x[ignuin(0, n - 1)];

  return(xStar);
}

VectorDouble bootstrapMB(const VectorDouble &x, int b, int nSame, int nStar)
// returns moving blocks bootstrap pseudo sequence with block size b
{
  int n = x.getSize(), l, i, j, index;

  if (nSame) nStar = n;
  VectorDouble xStar(nStar);
  l = nStar/b;

  for (i = 0; i < l; i++) {
    index = ignuin(0, n - b);
    for (j = 0; j < b; j++)
      xStar[i * b + j] = x[index + j];
  }
  index = ignuin(0, n - b);
  for (j = 0; j < (nStar - l * b); j++)
    xStar[l * b + j] = x[index + j];

  return(xStar);
}

VectorDouble bootstrapStat(const VectorDouble &x, int b, int nSame, int nStar)
// returns stationary bootstrap pseudo sequence with geometric parameter p = 1/b
{
  int n = x.getSize(), i, j, index;
  if (nSame) nStar = n;
  double p = 1.0 / ((double) b);
  VectorDouble xWrap = c(x, x), xStar(nStar + n);
  VectorInt bVec(1);

  // generate block sizes
  bVec[0] = min(gengeo(p), n);
  while (sum(bVec) < nStar)
    bVec = c(bVec, min(gengeo(p), n));

  index = ignuin(0, n - 1);
  for (j = 0; j < bVec[0]; j++)
    xStar[j] = xWrap[index + j];
  for (i = 1; i < bVec.getSize(); i++) {
    index = ignuin(0, n - 1);
    for (j = 0; j < bVec[i]; j++)
      xStar[sum(bVec(0, i - 1)) + j] = xWrap[index + j];
  }

  return xStar(0, nStar - 1);
}

MatrixDouble bootstrap(const MatrixDouble &x)
// returns iid bootstrap pseudo sequence (rows of x are observations)
{
  int n = x.getRow(), col = x.getCol(), i, j;
  VectorInt ind = sample(seq(0, n -1, 1), n, 1);
  MatrixDouble xStar(n, col);

  for (i = 0; i < n; i++)
    for (j = 0; j < col; j++)
       xStar(i, j) = x(ind[i], j);

  return xStar;
}

MatrixDouble bootstrapMB(const MatrixDouble &x, int b, int nSame, int nStar)
// returns multivariate moving blocks bootstrap pseudo sequence with block size b
{
  int n = x.getRow(), col = x.getCol(), l, i, j, k, index;

  if (nSame) nStar = n;
  MatrixDouble xStar(nStar, col);
  l = nStar/b;

  for (i = 0; i < l; i++) {
    index = ignuin(0, n - b);
    for (j = 0; j < b; j++)
      for (k = 0; k < col; k++)
        xStar(i * b + j, k) = x(index + j, k);
  }
  index = ignuin(0, n - b);
  for (j = 0; j < (nStar - l * b); j++)
    for (k = 0; k < col; k++)
      xStar(l * b + j, k) = x(index + j, k);

  return(xStar);
}

MatrixDouble bootstrapMBcircle(const MatrixDouble &x, int b, int nSame, int nStar)
// returns multivariate "circular" moving blocks bootstrap pseudo sequence
// with block size b (to avoid problem of recentering)
{
  int n = x.getRow(), col = x.getCol(), l, i, j, k, index;
  MatrixDouble xCircle = rbind(x, x);

  if (nSame) nStar = n;
  MatrixDouble xStar(nStar, col);
  l = nStar/b;

  for (i = 0; i < l; i++) {
    index = ignuin(0, n - 1);
    for (j = 0; j < b; j++)
      for (k = 0; k < col; k++)
        xStar(i * b + j, k) = xCircle(index + j, k);
  }
  index = ignuin(0, n - 1);
  for (j = 0; j < (nStar - l * b); j++)
    for (k = 0; k < col; k++)
      xStar(l * b + j, k) = xCircle(index + j, k);

  return(xStar);
}

MatrixDouble bootstrapStat(const MatrixDouble &x, int b, int nSame, int nStar)
// returns multivariate stationary bootstrap pseudo sequence with geometric parameter p = 1/b
{
  int n = x.getRow(), col = x.getCol(), i, j, k, index;
  if (nSame) nStar = n;
  double p = 1.0 / ((double) b);
  MatrixDouble xStar(nStar + n, col), xWrap = rbind(x, x);
  VectorInt bVec(1);

  // generate block sizes
  bVec[0] = min(gengeo(p), n);
  while (sum(bVec) < nStar) {
    bVec = c(bVec, min(gengeo(p), n));
  }

  index = ignuin(0, n - 1);
  for (j = 0; j < bVec[0]; j++)
    for (k = 0; k < col; k++)
      xStar(j, k) = xWrap(index + j, k);
  for (i = 1; i < bVec.getSize(); i++) {
    index = ignuin(0, n - 1);
    for (j = 0; j < bVec[i]; j++)
      for (k = 0; k < col; k++)
        xStar(sum(bVec(0, i - 1)) + j, k) = xWrap(index + j, k);
  }

  return extractRows(xStar, 0, nStar - 1);
}

MatrixDouble bootstrapConfReg(const MatrixDouble &x, const VectorDouble &y,
                      const VectorDouble &alpha, int betaInd, int addInt,
                      const VectorInt &seInd, int b, int w0, int addIntA,
                      MatrixDouble (*boot) (const MatrixDouble &, int, int, int),
                      int B)
/*
  - calculates bootstrap CIs for time series regression
  - both hybrid and studentized intervals are computed
  - both equal-tailed (row 1 and 2) and symmetric intervals (row 3 and 4)
    are computed
  - seInd specifies which (kernel) estimate of standard error to use to
    studentize the original statistic
  - if moving blocks bootstrap is used, we studentize the boostrap statistic
    along the lines of Gtze and Knsch (1996), if the stat. bootstrap is used
    we use the same kernel as for the original statistic
  - betaInd specifies the regression coefficient of interest
*/
{
  int n = x.getRow(), p = x.getCol(), n_alpha = alpha.getSize(), i, l = n / b,
      j, k;
  double betaHat, sigmaHat, quantile, seStar;
  const double eps = 0.0000001;
  VectorDouble fit1, Hybrid(B), Stud(B), yStar, zetaBar, zeta_j;
  MatrixDouble xy, xyStar, CI(4, 2 * n_alpha), fit, xStar, VStar(n, p + addInt),
               XTX, XTXnI, zeta(l, p + addInt), ShatStar;

  assert(n == y.getSize());

  xy = cbind(x, y);
  MatrixDouble xyAugmented(xy);
  if (addInt)
    xyAugmented = cbind(1, xy);

  fit = kernelCI(x, y, vec(alpha[0]), betaInd, addInt, 0, seInd, b, w0, addIntA);
  betaHat = fit(0, 0);
  sigmaHat = fit(0, 1);
  // check for positivity in case truncated kernel was used first
  if (sigmaHat < eps) {
    sigmaHat = fit(1, 1);
    cout << "Negative SD ";
  }

  for (i = 0; i < B; i++){
    //if (i % 100 == 0)

    // compute boostrap values according to whether MB or Stat Bootstrap is used
    if (boot == bootstrapMBcircle) {
      xyStar = (*boot)(xyAugmented, b, 1, 10);
      xStar = extractCols(xyStar, 0, p - 1 + addInt);
      yStar = extractCol(xyStar, p + addInt);
      fit = svdfitFull(xStar, yStar);
      XTX = nullMatrixDouble(p + addInt, p + addInt);
      for (j = 0; j < n; j++) {
        assignRow(VStar, j, extractRow(xStar, j) * fit(1, j));
        XTX += outerProduct(extractRow(xStar, j));
      }
      XTXnI = inv(XTX / (double) n);
      for (j = 0; j < l; j++) {
        zeta_j = nullVectorDouble(p + addInt);
        for (k = 0; k < b; k++)
          zeta_j += extractRow(VStar, j * b + k);
        zeta_j /= (double) b;
        assignRow(zeta, j, zeta_j);
      }
      zetaBar = apply(zeta, 2, mean);
      ShatStar = nullMatrixDouble(p + addInt, p + addInt);
      for (j = 0; j < l; j++)
        ShatStar += outerProduct(extractRow(zeta, j) - zetaBar);
      ShatStar *= ((double) b / (double) l);
      seStar = sqrt(max(eps, (XTXnI % ShatStar % XTXnI)(betaInd, betaInd)) / (double) n);
      Hybrid[i] = fit(0, betaInd) - betaHat;
      Stud[i] = Hybrid[i] / seStar;
    }
    else {
      xyStar = (*boot)(xy, b, 1, 10);
      fit = kernelCI(extractCols(xyStar, 0, p - 1), extractCol(xyStar, p),
                     vec(alpha[0]), betaInd, addInt, 0, seInd, w0, addIntA);
      Hybrid[i] = fit(0, 0) - betaHat;
      Stud[i] = Hybrid[i] / fit(0, 1);
    }
  }

  // compute equal-tailed intervals
  Hybrid = sort(Hybrid);
  Stud = sort(Stud);
  for (i = 0; i < n_alpha; i++) {
    CI(0, 2 * i) = betaHat - emp_quantile(1.0 - alpha[i]/2.0, Hybrid);
    CI(0, 2 * i + 1) = betaHat - emp_quantile(alpha[i]/2.0, Hybrid);
    CI(1, 2 * i) = betaHat - sigmaHat * emp_quantile(1.0 - alpha[i]/2.0, Stud);
    CI(1, 2 * i + 1) = betaHat - sigmaHat * emp_quantile(alpha[i]/2.0, Stud);
  }
  // compute symmetric intervals
  Hybrid = sort(fabs(Hybrid));
  Stud = sort(fabs(Stud));
  for (i = 0; i < n_alpha; i++) {
    quantile = emp_quantile(1.0 - alpha[i], Hybrid);
    CI(2, 2 * i) = betaHat - quantile;
    CI(2, 2 * i + 1) = betaHat + quantile;
    quantile = emp_quantile(1.0 - alpha[i], Stud);
    CI(3, 2 * i) = betaHat - sigmaHat * quantile;
    CI(3, 2 * i + 1) = betaHat + sigmaHat * quantile;
  }

  return CI;
}


MatrixDouble bootstrapConfReg(const VectorDouble &x, const VectorDouble &y,
                      const VectorDouble &alpha, int addInt,
                      const VectorInt &seInd, int b, int addIntA, int B,
                      MatrixDouble (*boot) (const MatrixDouble &, int, int, int))
/*
  - this is for simple linear regression!
  - calculates studentized bootstrap CIs for time series regression
  - for MB bootstrap both equal-tailed (row 1) and symmetric intervals
    (row 2) are computed
  - for Stat bootstrap only symmetric intervals are computed
  - if moving blocks bootstrap is used, we studentize the boostrap statistic
    along the lines of Gtze and Knsch (1996), if the stat. bootstrap is used
    we use the same kernel as for the original statistic
  - in the former case, seInd specifies rectangular kernel and alternative kernel,
    in the lattter case, seInd specifies the various kernels to be used
*/
{
  int n = x.getSize(), n_alpha = alpha.getSize(), i, l = n / b, j, k, MB = 0,
      n_se = seInd.getSize();
  double betaHat, sigmaHat, quantile, seStar;
  const double eps = 0.00000001;
  VectorDouble fit1, StudMB(B),  yStar, zetaBar, zeta_j, se;
  MatrixDouble xy, xyStar, CI(n_se, 2 * n_alpha), fit, xStar, VStar(n, 1 + addInt),
               XTX, XTXnI, zeta(l, 1 + addInt), ShatStar, StudStat(B, n_se);

  assert(n == y.getSize());

  if (boot == bootstrapMBcircle)
    MB = 1;

  xy = cbind(x, y);
  MatrixDouble xyAugmented(xy);
  if (addInt)
    xyAugmented = cbind(1, xy);

  fit = kernelCI(x, y, vec(alpha[0]), addInt, 0, seInd, b, addIntA);
  se = extractCol(fit, 1);
  betaHat = fit(0, 0);
  sigmaHat = fit(0, 1);
  // check for positivity in case truncated kernel was used first
  if (sigmaHat < eps) {
    sigmaHat = fit(1, 1);
    cout << "Negative SD ";
  }

  for (i = 0; i < B; i++){

    // compute boostrap values according to whether MB or Stat Bootstrap is used
    if (MB) {
      xyStar = (*boot)(xyAugmented, b, 1, 10);
      xStar = extractCols(xyStar, 0, addInt);
      yStar = extractCol(xyStar, 1 + addInt);
      fit = svdfitFull(xStar, yStar);
      XTX = nullMatrixDouble(1 + addInt, 1 + addInt);
      for (j = 0; j < n; j++) {
        assignRow(VStar, j, extractRow(xStar, j) * fit(1, j));
        XTX += outerProduct(extractRow(xStar, j));
      }
      XTXnI = inv(XTX / (double) n);
      for (j = 0; j < l; j++) {
        zeta_j = nullVectorDouble(1 + addInt);
        for (k = 0; k < b; k++)
          zeta_j += extractRow(VStar, j * b + k);
        zeta_j /= (double) b;
        assignRow(zeta, j, zeta_j);
      }
      zetaBar = apply(zeta, 2, mean);
      ShatStar = nullMatrixDouble(1 + addInt, 1 + addInt);
      for (j = 0; j < l; j++)
        ShatStar += outerProduct(extractRow(zeta, j) - zetaBar);
      ShatStar *= ((double) b / (double) l);
      seStar = sqrt(max(eps, (XTXnI % ShatStar % XTXnI)(addInt, addInt)) / (double) n);
      StudMB[i] = (fit(0, addInt) - betaHat) / seStar;
    }

    // stationary bootstrap
    else {
      xyStar = (*boot)(xy, b, 1, 10);
      fit = kernelCI(extractCol(xyStar, 0), extractCol(xyStar, 1),
                     vec(alpha[0]), addInt, 0, seInd, addIntA);
      for (k = 0; k < n_se; k++)
        StudStat(i, k) = (fit(0, 0) - betaHat) / fit(k, 1);
    }
  }

  // final intervals according to MB or not
  if (MB) {
    // equal-tailed interval
    StudMB = sort(StudMB);
    for (i = 0; i < n_alpha; i++) {
      CI(0, 2 * i) = betaHat - sigmaHat * emp_quantile(1.0 - alpha[i]/2.0, StudMB);
      CI(0, 2 * i + 1) = betaHat - sigmaHat * emp_quantile(alpha[i]/2.0, StudMB);
    }
    // symmetric interval
    StudMB = sort(fabs(StudMB));
    for (i = 0; i < n_alpha; i++) {
      quantile = emp_quantile(1.0 - alpha[i], StudMB);
      CI(1, 2 * i) = betaHat - sigmaHat * quantile;
      CI(1, 2 * i + 1) = betaHat + sigmaHat * quantile;
    }
  }
  else {
    for (k = 0; k < n_se; k++) {
      StudMB = sort(fabs(extractCol(StudStat, k)));  // abuse of "notation"
      for (i = 0; i < n_alpha; i++) {
        quantile = emp_quantile(1.0 - alpha[i], StudMB);
        CI(k, 2 * i) = betaHat - se[k] * quantile;
        CI(k, 2 * i + 1) = betaHat + se[k] * quantile;
      }
    }
  }

  return CI;
}


MatrixDouble bootstrapConfRegFast(const VectorDouble &x, const VectorDouble &y,
                      const VectorDouble &alpha, int addInt,
                      const VectorInt &seInd, int b, int addIntA, int B,
                      MatrixDouble (*boot) (const MatrixDouble &, int, int, int))
/*
  - this is for simple linear regression!
  - calculates studentized bootstrap CIs for time series regression
  - for MB bootstrap both equal-tailed (row 1) and symmetric intervals
    (row 2) are computed; in addition computes corresponding hybrid
    intervals (row 3 and 4, repectively)
  - for Stat bootstrap only symmetric intervals are computed
  - if moving blocks bootstrap is used, we studentize the boostrap statistic
    along the lines of Gtze and Knsch (1996), if the stat. bootstrap is used
    we use the same kernel as for the original statistic
  - in the former case, seInd specifies rectangular kernel and alternative kernel,
    in the lattter case, seInd specifies the various kernels to be used
*/
{ int n = x.getSize(), n_alpha = alpha.getSize(), i, l = n / b, j, k,
      n_se = seInd.getSize(), MB = (boot == bootstrapMBcircle);
  double betaHat, sigmaHat, quantile, seStar, zeta_j, zetaBar, XTX, XTXnI,
         ShatStar, betaStar;
  const double eps = 0.00000001;
  VectorDouble fit1, StudMB(B),  Hybrid(B), xStar, yStar, se, VStar(n), zeta(l);
  MatrixDouble xy, xyStar, CI(MB * 4 + (1 - MB) * n_se, 2 * n_alpha), fit,
               StudStat(B, n_se);

  assert(n == y.getSize());

  xy = cbind(x, y);

  fit = kernelCI(x, y, vec(alpha[0]), addInt, 0, seInd, b, addIntA);
  se = extractCol(fit, 1);
  betaHat = fit(0, 0);
  sigmaHat = fit(0, 1);
  // check for positivity in case truncated kernel was used first
  if (sigmaHat < eps) {
    sigmaHat = fit(1, 1);
    cout << "Negative SD ";
  }

  for (i = 0; i < B; i++){

    // compute boostrap values according to whether MB or Stat Bootstrap is used
    if (MB) {
      xyStar = (*boot)(xy, b, 1, 10);
      xStar = extractCol(xyStar, 0);
      yStar = extractCol(xyStar, 1);
      if (addInt) {
        xStar -= mean(xStar);
        yStar -= mean(yStar);
      }
      XTX = innerProduct(xStar, xStar);
      betaStar = innerProduct(xStar, yStar) / XTX;
      VStar = xStar * (yStar - xStar * betaStar);
      XTXnI = ((double) n / XTX);

      for (j = 0; j < l; j++) {
        zeta_j = 0;
        for (k = 0; k < b; k++)
          zeta_j += VStar[j * b + k];
        zeta_j /= (double) b;
        zeta[j] = zeta_j;
      }
      zetaBar = mean(zeta);
      ShatStar = 0;
      for (j = 0; j < l; j++)
        ShatStar += pow(zeta[j] - zetaBar, 2);
      ShatStar *= ((double) b / (double) l);
      seStar = sqrt(max(eps, (XTXnI * ShatStar * XTXnI)) / (double) n);
      Hybrid[i] = betaStar - betaHat;
      StudMB[i] = Hybrid[i] / seStar;
    }

    // stationary bootstrap
    else {
      xyStar = (*boot)(xy, b, 1, 10);
      fit = kernelCI(extractCol(xyStar, 0), extractCol(xyStar, 1),
                     vec(alpha[0]), addInt, 0, seInd, addIntA);
      for (k = 0; k < n_se; k++)
        StudStat(i, k) = (fit(0, 0) - betaHat) / fit(k, 1);
    }
  }

  // final intervals according to MB or not
  if (MB) {
    // equal-tailed interval
    StudMB = sort(StudMB);
    for (i = 0; i < n_alpha; i++) {
      CI(0, 2 * i) = betaHat - sigmaHat * emp_quantile(1.0 - alpha[i]/2.0, StudMB);
      CI(0, 2 * i + 1) = betaHat - sigmaHat * emp_quantile(alpha[i]/2.0, StudMB);
    }
    // symmetric interval
    StudMB = sort(fabs(StudMB));
    for (i = 0; i < n_alpha; i++) {
      quantile = emp_quantile(1.0 - alpha[i], StudMB);
      CI(1, 2 * i) = betaHat - sigmaHat * quantile;
      CI(1, 2 * i + 1) = betaHat + sigmaHat * quantile;
    }
    // equal-tailed hybrid interval
    Hybrid = sort(Hybrid);
    for (i = 0; i < n_alpha; i++) {
      CI(2, 2 * i) = betaHat - emp_quantile(1.0 - alpha[i]/2.0, Hybrid);
      CI(2, 2 * i + 1) = betaHat - emp_quantile(alpha[i]/2.0, Hybrid);
    }
    // symmetric interval
    Hybrid = sort(fabs(Hybrid));
    for (i = 0; i < n_alpha; i++) {
      quantile = emp_quantile(1.0 - alpha[i], Hybrid);
      CI(3, 2 * i) = betaHat - quantile;
      CI(3, 2 * i + 1) = betaHat + quantile;
    }
  }
  else {
    for (k = 0; k < n_se; k++) {
      StudMB = sort(fabs(extractCol(StudStat, k)));  // abuse of "notation"
      for (i = 0; i < n_alpha; i++) {
        quantile = emp_quantile(1.0 - alpha[i], StudMB);
        CI(k, 2 * i) = betaHat - se[k] * quantile;
        CI(k, 2 * i + 1) = betaHat + se[k] * quantile;
      }
    }
  }
  return CI;
}


int stopCheck(const MatrixDouble &Pvalues, double alphaMax)
// for functions below, checks whether we can stop
{
  VectorDouble p(4);

  p[0] = min(extractRow(Pvalues, 0));
  p[0] = 2 * min(p[0], 1 - p[0]);
  p[1] = min(extractRow(Pvalues, 1));
  p[2] = min(extractRow(Pvalues, 2));
  p[2] = 2 * min(p[2], 1 - p[2]);
  p[3] = min(extractRow(Pvalues, 3));

  if (min(p) >= alphaMax)
    return 1;
  else
    return 0;
}


MatrixDouble bootstrapPvalRegFast(const VectorDouble &x, const VectorDouble &y,
                      double alphaMax, double beta0, int addInt,
                      const VectorInt &seInd, int b, int addIntA,
                      MatrixDouble (*boot) (const MatrixDouble &, int, int, int),
                      int GK, int B)
/*
  - this is for simple linear regression!
  - calculates bootstrap Pvalues for time series regression
  - both hybrid and studentized P-values are computed
  - both equal-tailed and symmetric P-values are computed
  - (row) order is: noStud-ET, noStud-SYM, Stud-ET, Stud-Sym
  - seInd specifies which (kernel) estimate of standard error to use to
    studentize the original statistic
  - our method uses the same kernel the studentize both the original and
    the bootstrap statistic
  - if the moving blocks bootstrap is used,
    also do the boostrap studentization along the lines of
    Gtze and Knsch (1996), if the stat. bootstrap is used
    we use the same kernel as for the original statistic
  - seInd specifies which kernel(s) to use; if GK = 1, then the second
    element of seInd should correspond to the truncated kernel, the first
    element must be a kernel with characteristic exponent q <= 2 to ensure
    a positive standard error
*/
{ int n = x.getSize(), i, l = n / b, j, k, MB = (boot == bootstrapMBcircle);
  double betaHat, betaHatStar, se, seTrunc, seStar, zeta_j, zetaBar,
         XTX, XTXnI, ShatStar, Binv = 1.0 / (double) B;
  const double eps = 0.000001;
  VectorDouble fit1, xStar, yStar, VStar(n), zeta(l), hybrid(B);
  MatrixDouble xy, xyStar, fit, Pvalues = nullMatrixDouble(5, 1 + MB);

  assert(n == y.getSize());

  xy = cbind(x, y);

  fit = kernelCI(x, y, vec(0.05), addInt, 0, seInd, b, addIntA);
  betaHat = fit(0, 0);
  se = fit(0, 1);

  if (1 == GK) {
    seTrunc = fit(1, 1);
    if (seTrunc < eps) {
      Pvalues(4, 0) = 1;
      seTrunc = se;
      // cout << "Negative SD for b = " << b << ", ";
    }
  }

  for (i = 0; i < B; i++){

    // compute boostrap values according to whether MB or Stat Bootstrap is used

    if (boot == bootstrapMBcircle) {
      xyStar = (*boot)(xy, b, 1, 10);
      xStar = extractCol(xyStar, 0);
      yStar = extractCol(xyStar, 1);
      fit = kernelCI(xStar, yStar, vec(0.05), addInt, 0, seInd, b, addIntA);
      betaHatStar = fit(0, 0);
      seStar = fit(0, 1);

      // noStud
      if ((betaHatStar - betaHat) > (betaHat - beta0))
        Pvalues(0, 0) += Binv;
      if (fabs(betaHatStar - betaHat) > fabs(betaHat - beta0))
        Pvalues(1, 0) += Binv;
      // Stud
      if ((betaHatStar - betaHat)/seStar > (betaHat - beta0)/se)
        Pvalues(2, 0) += Binv;
      if (fabs(betaHatStar - betaHat)/seStar > fabs(betaHat - beta0)/se)
        Pvalues(3, 0) += Binv;

      if (1 == GK) {
        if (addInt) {
          xStar -= mean(xStar);
          yStar -= mean(yStar);
        }
        XTX = innerProduct(xStar, xStar);
        if (fabs(innerProduct(xStar, yStar) / XTX - betaHatStar) > 0.00001)
          cout << "Shit! ";
        VStar = xStar * (yStar - xStar * betaHatStar);
        XTXnI = ((double) n / XTX);

        for (j = 0; j < l; j++) {
          zeta_j = 0;
          for (k = 0; k < b; k++)
            zeta_j += VStar[j * b + k];
            zeta_j /= (double) b;
          zeta[j] = zeta_j;
        }
        zetaBar = mean(zeta);
        ShatStar = 0;
        for (j = 0; j < l; j++)
          ShatStar += pow(zeta[j] - zetaBar, 2);
        ShatStar *= ((double) b / (double) l);
        if (XTXnI * ShatStar * XTXnI > eps)
          seStar = sqrt((XTXnI * ShatStar * XTXnI) / (double) n);
        else {
          // cout << "Negative Stud-SD ";
          seStar = eps;
        }
        // noStud
        if ((betaHatStar - betaHat) > (betaHat - beta0))
          Pvalues(0, 1) += Binv;
        if (fabs(betaHatStar - betaHat) > fabs(betaHat - beta0))
          Pvalues(1, 1) += Binv;
        // Stud
        if ((betaHatStar - betaHat)/seStar > (betaHat - beta0)/seTrunc)
          Pvalues(2, 1) += Binv;
        if (fabs(betaHatStar - betaHat)/seStar > fabs(betaHat - beta0)/seTrunc)
          Pvalues(3, 1) += Binv;
      }
    }

    // stationary bootstrap
    else {
      xyStar = (*boot)(xy, b, 1, 10);
      fit = kernelCI(extractCol(xyStar, 0), extractCol(xyStar, 1),
                     vec(0.05), addInt, 0, seInd, addIntA);
      betaHatStar = fit(0, 0);
      seStar = fit(0, 1);

      // noStud
      if ((betaHatStar - betaHat) > (betaHat - beta0))
        Pvalues(0, 0) += Binv;
      if (fabs(betaHatStar - betaHat) > fabs(betaHat - beta0))
        Pvalues(1, 0) += Binv;
      // Stud
      if ((betaHatStar - betaHat)/seStar > (betaHat - beta0)/se)
        Pvalues(2, 0) += Binv;
      if (fabs(betaHatStar - betaHat)/seStar > fabs(betaHat - beta0)/se)
        Pvalues(3, 0) += Binv;

      /* can fill this in later ....
      if (1 == GK) {
        copy the above .....
      }  */

    }
    // see if we can stop
    if (1 == stopCheck(Pvalues, alphaMax))
      i = B;
  }

  for (j = 0; j < 1 + GK; j++) {
    Pvalues(0, j) = 2 * min(c(Pvalues(0, j), 1 - Pvalues(0, j)));
    Pvalues(2, j) = 2 * min(c(Pvalues(2, j), 1 - Pvalues(2, j)));
  }

  return Pvalues;
}


MatrixDouble bootstrapPvalReg(const MatrixDouble &x, const VectorDouble &y,
                      double alphaMax, int betaInd, double beta0, int addInt,
                      const VectorInt &seInd, int b, int svd, int w0, int addIntA,
                      MatrixDouble (*boot) (const MatrixDouble &, int, int, int),
                      int GK, int B)
/*
  - bootstrap analysis for time series regression
  - checks whether P-values are above alphaMax (say 0.1)
  - if yes, calculation is stopped
  - idea: this should save a lot of time under the null!
  - both hybrid and studentized P-values are computed
  - both equal-tailed and symmetric P-values are computed
  - (row) order is: noStud-ET, noStud-SYM, Stud-ET, Stud-Sym
  - seInd specifies which (kernel) estimate of standard error to use to
    studentize the original statistic
  - our method uses the same kernel the studentize both the original and
    the bootstrap statistic
  - if GK = 1, also do the boostrap studentization along the lines of
    Gtze and Knsch (1996), if the stat. bootstrap is used
    we use the same kernel as for the original statistic
  - seInd specifies which kernel(s) to use; if GK = 1, then the second
    element of seInd should correspond to the truncated kernel, the first
    element must be a kernel with characteristic exponent q <= 2 to ensure
    a positive standard error
*/
{
  int n = x.getRow(), p = x.getCol(), i, j, k, l = n / b;
  double betaHat, se, seTrunc, betaHatStar, seStar;
  const double eps = 0.0000001, Binv = 1 / (double) B;
  VectorDouble yStar, zetaBar, zeta_j;
  MatrixDouble xy, xyStar, Pvalues = nullMatrixDouble(5, 1 + GK), fit, xStar,
               VStar(n, p + addInt), XTX, XTXnI, zeta(l, p + addInt), ShatStar;

  assert(n == y.getSize());

  xy = cbind(x, y);
  MatrixDouble xyAugmented(xy);
  if (addInt)
    xyAugmented = cbind(1, xy);

  fit = kernelCI(x, y, vec(alphaMax), betaInd, addInt, beta0, seInd, svd, b, w0, addIntA);
  betaHat = fit(0, 0);
  se = fit(0, 1);

  if (1 == GK) {
    seTrunc = fit(1, 1);
    if (seTrunc < eps) {
      Pvalues(4, 0) = 1;
      seTrunc = se;
      cout << "Negative SD for b = " << b << ", ";
    }
  }

  for (i = 0; i < B; i++){

    // compute boostrap values according to whether MB or Stat Bootstrap is used

    if (boot == bootstrapMBcircle) {
      xyStar = (*boot)(xyAugmented, b, 1, 10);
      xStar = extractCols(xyStar, 0, p - 1 + addInt);
      yStar = extractCol(xyStar, p + addInt);
      fit = kernelCI(xStar, yStar, vec(alphaMax), betaInd, 0, beta0,
                     vec(seInd[0]), svd, b, w0, addIntA);
      betaHatStar = fit(0, 0);
      seStar = fit(0, 1);

      // noStud
      if ((betaHatStar - betaHat) > (betaHat - beta0))
        Pvalues(0, 0) += Binv;
      if (fabs(betaHatStar - betaHat) > fabs(betaHat - beta0))
        Pvalues(1, 0) += Binv;
      // Stud
      if ((betaHatStar - betaHat)/seStar > (betaHat - beta0)/se)
        Pvalues(2, 0) += Binv;
      if (fabs(betaHatStar - betaHat)/seStar > fabs(betaHat - beta0)/se)
        Pvalues(3, 0) += Binv;

      if (1 == GK) {
        fit = svdfitFull(xStar, yStar);
        XTX = nullMatrixDouble(p + addInt, p + addInt);
        for (j = 0; j < n; j++) {
          assignRow(VStar, j, extractRow(xStar, j) * fit(1, j));
          XTX += outerProduct(extractRow(xStar, j));
        }
        XTXnI = inv(XTX / (double) n);
        for (j = 0; j < l; j++) {
          zeta_j = nullVectorDouble(p + addInt);
          for (k = 0; k < b; k++)
            zeta_j += extractRow(VStar, j * b + k);
          zeta_j /= (double) b;
          assignRow(zeta, j, zeta_j);
        }
        zetaBar = apply(zeta, 2, mean);
        ShatStar = nullMatrixDouble(p + addInt, p + addInt);
        for (j = 0; j < l; j++)
          ShatStar += outerProduct(extractRow(zeta, j) - zetaBar);
        ShatStar *= ((double) b / (double) l);
        if ((XTXnI % ShatStar % XTXnI)(betaInd, betaInd) > eps)
          seStar = sqrt((XTXnI % ShatStar % XTXnI)(betaInd, betaInd) / (double) n);
        else
          cout << "Negative Boot-SD ";

        // noStud
        if ((betaHatStar - betaHat) > (betaHat - beta0))
          Pvalues(0, 1) += Binv;
        if (fabs(betaHatStar - betaHat) > fabs(betaHat - beta0))
          Pvalues(1, 1) += Binv;
        // Stud
        if ((betaHatStar - betaHat)/seStar > (betaHat - beta0)/seTrunc)
          Pvalues(2, 1) += Binv;
        if (fabs(betaHatStar - betaHat)/seStar > fabs(betaHat - beta0)/seTrunc)
          Pvalues(3, 1) += Binv;
      }
    }
    else {
      xyStar = (*boot)(xyAugmented, b, 1, 10);
      xStar = extractCols(xyStar, 0, p - 1 + addInt);
      yStar = extractCol(xyStar, p + addInt);
      fit = kernelCI(xStar, yStar, vec(alphaMax), betaInd, addInt, beta0,
                     vec(seInd[0]), svd, b, w0, addIntA);
      betaHatStar = fit(0, 0);
      seStar = fit(0, 1);

      // noStud
      if ((betaHatStar - betaHat) > (betaHat - beta0))
        Pvalues(0, 0) += Binv;
      if (fabs(betaHatStar - betaHat) > fabs(betaHat - beta0))
        Pvalues(1, 0) += Binv;
      // Stud
      if ((betaHatStar - betaHat)/seStar > (betaHat - beta0)/se)
        Pvalues(2, 0) += Binv;
      if (fabs(betaHatStar - betaHat)/seStar > fabs(betaHat - beta0)/se)
        Pvalues(3, 0) += Binv;

      if (1 ==GK) {
        xyStar = bootstrapMBcircle(xyAugmented, b);
        xStar = extractCols(xyStar, 0, p - 1 + addInt);
        yStar = extractCol(xyStar, p + addInt);
        fit = svdfitFull(xStar, yStar);
        betaHatStar = fit(0, betaInd);
        XTX = nullMatrixDouble(p + addInt, p + addInt);
        for (j = 0; j < n; j++) {
          assignRow(VStar, j, extractRow(xStar, j) * fit(1, j));
          XTX += outerProduct(extractRow(xStar, j));
        }
        XTXnI = inv(XTX / (double) n);
        for (j = 0; j < l; j++) {
          zeta_j = nullVectorDouble(p + addInt);
          for (k = 0; k < b; k++)
            zeta_j += extractRow(VStar, j * b + k);
          zeta_j /= (double) b;
          assignRow(zeta, j, zeta_j);
        }
        zetaBar = apply(zeta, 2, mean);
        ShatStar = nullMatrixDouble(p + addInt, p + addInt);
        for (j = 0; j < l; j++)
          ShatStar += outerProduct(extractRow(zeta, j) - zetaBar);
        ShatStar *= ((double) b / (double) l);
        seStar = sqrt(max(eps, (XTXnI % ShatStar % XTXnI)(betaInd, betaInd)) / (double) n);

        // noStud
        if ((betaHatStar - betaHat) > (betaHat - beta0))
          Pvalues(0, 1) += Binv;
        if (fabs(betaHatStar - betaHat) > fabs(betaHat - beta0))
          Pvalues(1, 1) += Binv;
        // Stud
        if ((betaHatStar - betaHat)/seStar > (betaHat - beta0)/seTrunc)
          Pvalues(2, 1) += Binv;
        if (fabs(betaHatStar - betaHat)/seStar > fabs(betaHat - beta0)/seTrunc)
          Pvalues(3, 1) += Binv;
      }
    }
    // see if we can stop
    if (1 == stopCheck(Pvalues, alphaMax))
      i = B;
  }

  for (j = 0; j < 1 + GK; j++) {
    Pvalues(0, j) = 2 * min(c(Pvalues(0, j), 1 - Pvalues(0, j)));
    Pvalues(2, j) = 2 * min(c(Pvalues(2, j), 1 - Pvalues(2, j)));
  }

  return Pvalues;
}


