/*****

File: boot.h
--------------
Library package with bootstrap routines
for self-contained use in C++ rather than export to Splus

*****/

#ifndef _boot_h
#define _boot_h

#include "vector.h"
#include "matrix.h"

VectorDouble bootstrap(const VectorDouble &x, int nSame = 1, int nStar = 10);
VectorDouble bootstrapMB(const VectorDouble &x, int b, int nSame = 1,
                         int nStar = 10);
VectorDouble bootstrapStat(const VectorDouble &x, int b, int nSame = 1,
                           int nStar = 10);
MatrixDouble bootstrap(const MatrixDouble &x);
MatrixDouble bootstrapMB(const MatrixDouble &x, int b, int nSame = 1,
                         int nStar = 10);
MatrixDouble bootstrapMBcircle(const MatrixDouble &x, int b, int nSame = 1,
                               int nStar = 10);
MatrixDouble bootstrapStat(const MatrixDouble &x, int b, int nSame = 1,
                           int nStar = 10);

MatrixDouble bootstrapConfReg(const MatrixDouble &x, const VectorDouble &y,
                      const VectorDouble &alpha, int betaInd, int addInt,
                      const VectorInt &seInd, int b, int w0 = 0, int addIntA = 1,
                      MatrixDouble (*boot) (const MatrixDouble &, int, int, int)
                      = bootstrapStat, int B = 100);

MatrixDouble bootstrapConfReg(const VectorDouble &x, const VectorDouble &y,
                      const VectorDouble &alpha, int addInt,
                      const VectorInt &seInd, int b, int addIntA, int B,
                      MatrixDouble (*boot) (const MatrixDouble &, int, int, int)
                      = bootstrapStat);

MatrixDouble bootstrapConfRegFast(const VectorDouble &x, const VectorDouble &y,
                      const VectorDouble &alpha, int addInt,
                      const VectorInt &seInd, int b, int addIntA, int B,
                      MatrixDouble (*boot) (const MatrixDouble &, int, int, int)
                      = bootstrapStat);

MatrixDouble bootstrapPvalRegFast(const VectorDouble &x, const VectorDouble &y,
                      double alphaMax, double beta0, int addInt,
                      const VectorInt &seInd, int b, int addIntA = 0,
                      MatrixDouble (*boot) (const MatrixDouble &, int, int, int)
                      = bootstrapMBcircle, int GK = 1, int B = 1000);

MatrixDouble bootstrapPvalReg(const MatrixDouble &x, const VectorDouble &y,
                      double alphaMax, int betaInd, double beta0, int addInt,
                      const VectorInt &seInd, int b, int svd = 1, int w0 = 0,
                      int addIntA = 0,
                      MatrixDouble (*boot) (const MatrixDouble &, int, int, int)
                      = bootstrapMBcircle, int GK = 1, int B = 1000);

#endif /* boot.h */


