
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <assert.h>

#include "statfcts.h" 
#include "nrutil.h"
#include "sort.h"
#include "vector.h"
#include "matrix.h"
#include "math.h"


int max(int a, int b)
{
  if (a > b)
    return a;
  else
    return b;
}

int min(int a, int b)
{
  if (a > b)
    return b;
  else
    return a;
}

double max(double a, double b)
{
  if (a > b)
    return a;
  else
    return b;
}

double min(double a, double b)
{
  if (a > b)
    return b;
  else
    return a;
}

int min(const VectorInt &x)
{
  int min;

  min = x[0];

  for (int i = 1; i < x.getSize(); i++)
    if (x[i] < min)
      min = x[i];

  return(min);
}

int max(const VectorInt &x)
{
  int max;

  max = x[0];

  for (int i = 1; i < x.getSize(); i++)
    if (x[i] > max)
      max = x[i];

  return(max);
}

double min(const VectorDouble &x)
{
  double min;

  min = x[0];

  for (int i = 1; i < x.getSize(); i++)
    if (x[i] < min)
      min = x[i];

  return(min);
}

double max(const VectorDouble &x)
{
  double max;

  max = x[0];

  for (int i = 1; i < x.getSize(); i++)
    if (x[i] > max)
      max = x[i];

  return(max);
}

VectorInt seq(int from, int to, int by)
// mimicks Splus function for int sequences
{
  int length = (to - from) / by + 1;
  VectorInt sequence(length);

  sequence[0] = from;
  for (int i = 1; i < length; i++)
    sequence[i] = sequence[i - 1] + by;

  return(sequence);
}

VectorDouble seq(double from, double to, double by)
// mimicks Splus function for double sequences
{
  int length = (int) ((to - from) / by + 1);
  VectorDouble sequence(length);

  sequence[0] = from;
  for (int i = 1; i < length; i++)
    sequence[i] = sequence[i - 1] + by;

  return(sequence);
}

int contained(int x, const VectorInt &vec)
// checks whether x is contained in vec
{
  for (int i = 0; i < vec.getSize(); i++)
    if (x == vec[i]) {
      return 1;
    }

  return 0;
}

VectorDouble fabs(const VectorDouble &x)
{
  VectorDouble fabsX(x);

  for (int i = 0; i < fabsX.getSize(); i++)
    fabsX[i] = fabs(x[i]);

  return(fabsX);
}

VectorDouble pow(const VectorDouble &x, double power)
{
  VectorDouble powX(x);

  for (int i = 0; i < powX.getSize(); i++)
    powX[i] = pow(x[i], power);

  return(powX);
}

VectorDouble powProb(const VectorDouble &x, double power)
{
  VectorDouble powX(x);

  for (int i = 0; i < powX.getSize(); i++)
    if (x[i] > 0)
      powX[i] = pow(x[i], power);
    else
      powX[i] = 0;

  return(powX);
}

VectorDouble log(const VectorDouble &x)
{
  VectorDouble logX(x);

  for (int i = 0; i < logX.getSize(); i++)
    logX[i] = log(x[i]);

  return logX;
}

VectorDouble exp(const VectorDouble &x)
{
  VectorDouble expX(x);

  for (int i = 0; i < expX.getSize(); i++)
    expX[i] = exp(x[i]);

  return expX;
}

MatrixDouble center(const MatrixDouble &x, int whichWay)
// subtract means either row- or columnwise
{
  int i, j, n = x.getRow(), p = x.getCol();
  VectorDouble xMeans;
  MatrixDouble xCenter(n, p);

  xMeans = apply(x, whichWay, mean);

  if (whichWay == 2) {
    for (i = 0; i < n; i++)
      for (j = 0; j < p; j++)
        xCenter(i, j) = x(i, j) - xMeans[j];
  }
  else {
    for (i = 0; i < n; i++)
      for (j = 0; j < p; j++)
        xCenter(i, j) = x(i, j) - xMeans[i];
  }

  return xCenter;
}

MatrixDouble outerProduct(const VectorDouble &x)
// returns outer product of x
{
  int n = x.getSize();
  MatrixDouble prod(n, n);

  for (int i = 0; i < n; i++)
    for (int j = 0; j < n; j++)
      prod(i, j) = x[i] * x[j];

  return prod;
}

MatrixDouble outerProduct(const VectorDouble &x, const VectorDouble &y)
// returns outer product of x
{
  int n = x.getSize();
  MatrixDouble prod(n, n);

  assert(n == y.getSize());

  for (int i = 0; i < n; i++)
    for (int j = 0; j < n; j++)
      prod(i, j) = x[i] * y[j];

  return prod;
}

double euclidNorm(const VectorDouble &x, int standardize)
// compute Euclidian norm (potentially standardized)
{
  int i, n = x.getSize();
  double norm2 = 0;

  for (i = 0; i < n; i++)
    norm2 += pow(x[i], 2);

  if (standardize)
    return sqrt(norm2 / (double) n);
  else
    return sqrt(norm2);
}

double euclidNormSquare(const VectorDouble &x, int standardize)
// compute Euclidian norm (potentially standardized)
{
  int i, n = x.getSize();
  double norm2 = 0;

  for (i = 0; i < n; i++)
    norm2 += pow(x[i], 2);

  if (standardize)
    return norm2 / (double) n;
  else
    return norm2;
}

double supNorm(const VectorDouble &x)
// compute sup Norm
{
  return max(fabs(x));
}

VectorDouble Delta(const VectorDouble &x, int times)
// returns 'times' differenced vector
{
  int n = x.getSize();

  assert(times >= 1);
  assert(times < n);

  if (times == 1)
    return c(0, x(1, n - 1) - x(0, n - 2));
  else
    return Delta(c(x[1] - x[0], x(1, n - 1) - x(0, n - 2)), times - 1);
}

MatrixDouble Delta(const MatrixDouble &x, int rowWise, int times)
// returns 'times' differenced matrix, row- or columnwise
{
  int row = x.getRow(), col = x.getCol();

  assert(times >= 1);
  if (rowWise)
    assert(times < row);
  else
    assert(times < col);

  if (rowWise) {
    if (times == 1)
      return rbind(nullMatrixDouble(1, col),
                   extractRows(x, 1, row - 1) - extractRows(x, 0, row - 2));
    else
      return Delta(rbind(extractRows(x, 1, 1) - extractRows(x, 0, 0),
                         extractRows(x, 1, row - 1) - extractRows(x, 0, row - 2)),
                   times - 1);
  }
  else {
    if (times == 1)
      return cbind(nullVectorDouble(row),
                   extractCols(x, 1, col - 1) - extractCols(x, 0, col - 2));
    else
      return Delta(cbind(extractCol(x, 1) - extractCol(x, 0),
                         extractCols(x, 1, col - 1) - extractCols(x, 0, col - 2)),
                   times - 1);
  }
}

VectorDouble null2(double a, double b, double c)
// computes null points of ax^2 + bx + c (in ascending order)
{
  assert(a != 0 && b != 0);

  double eps = 0.0000000001, help;
  VectorDouble sols(2);

  if (fabs(a) < eps)
    sols[0] = sols[1] = -c / b;
  else {
    assert(pow(b, 2) - 4 * a * c >= 0);
    sols[0] = (-b + sqrt(pow(b, 2) - 4 * a * c)) / (2 * a);
    help = (-b - sqrt(pow(b, 2) - 4 * a * c)) / (2 * a);
    if (help < sols[0]) {
      sols[1] = sols[0];
      sols[0] = help;
    }
    else
      sols[1] = help;
  }

  return sols;
}

MatrixDouble eigen2(const MatrixDouble &x)
/*
- returns eigenvalues and eigenvectors of 2 x 2 matrix
- first column contains the eigenvalues in descending order
- next two columns contain the corresponding eigenvectors
*/
{
  assert(x.getRow() == 2 && x.getCol() == 2);

  VectorDouble nullPoints;
  MatrixDouble result(2, 3);

  // compute null points of characteristic polynomial (in ascending order)
  nullPoints = null2(1, - x(0, 0) - x(1, 1), x(0, 0) * x(1, 1) - x(0, 1) * x(1, 0));
  result(0, 0) = nullPoints[1];
  result(1, 0) = nullPoints[0];

  // compute corresponding eigenvectors (as null points of CP)
  result(0, 1) = - x(0, 1) / (x(0, 0) - nullPoints[1]);
  result(1, 1) = 1;

  result(0, 2) = 1;
  result(1, 2) = - x(1, 0) / (x(1, 1) - nullPoints[0]);

  return result;
}

double trace(const MatrixDouble &x)
// computes, yes, the trace
{
  int i, col = x.getCol();
  double tr = 0;

  assert(col == x.getRow());

  for (i = 0; i < col; i++)
    tr += x[i][i];

  return tr;
}

/*
does not seem to work !!!!

double det(const MatrixDouble &x, int takeLogs)
// computes determinant of a matrix based on LU decomposition
// if takeLogs == 1, logs are taking to fight overflow
{
  int col = x.getCol(), i, j, sign = 1;
  int *indx;
  double **xmat, d = 1, logDet = 0;
  const double tol = 1.0e-16;

  assert(col == x.getRow());

  xmat = dmatrix(1, col, 1, col);
  indx = ivector(1, col);
  for (i = 0; i < col; i++)
    for (j = 0; j < col; j++)
      xmat[i + 1][j + 1] = x[i][j];

  ludcmp(xmat, col, indx, &d);

  if (0 == takeLogs) {
    for (i = 1; i<= col; i++) {
    //cout << xmat[i][i] << " ";
      if (fabs(xmat[i][i]) < tol) {
        free_dmatrix(xmat, 1, col, 1, col);
        free_ivector(indx, 1, col);
        return 0.0;
      }
      d *= xmat[i][i];
    }
    //cout << endl;
    free_dmatrix(xmat, 1, col, 1, col);
    free_ivector(indx, 1, col);
    return d;
  }

  else {
    for (i = 1; i <= col; i++) {
      if (fabs(xmat[i][i]) < tol) {
        free_dmatrix(xmat, 1, col, 1, col);
        free_ivector(indx, 1, col);
        return 0.0;
      }
      logDet += log(fabs(xmat[i][i]));
      if (xmat[i][i] < 0)
        sign *= -1;
    }
    free_dmatrix(xmat, 1, col, 1, col);
    free_ivector(indx, 1, col);
    return sign * d * exp(logDet);
  }
}
*/

double frobNorm(const MatrixDouble &x, int standardize)
// compute Frobenius norm of matrix x
// if standardize == 1 divide by dimensionality
{
  int i, j, row = x.getRow(), col = x.getCol();
  double norm2 = 0, dim = row * col;

  for (i = 0; i < row; i++)
    for (j = 0; j < col; j++)
      norm2 += pow(x(i, j), 2);

  if (standardize)
    return sqrt(norm2 / dim);
  else
    return sqrt(norm2);
}

double frobNormSquare(const MatrixDouble &x, int standardize)
// compute squared Frobenius norm of matrix x
// if standardize == 1 divide by dimensionality
{
  int i, j, row = x.getRow(), col = x.getCol();
  double norm2 = 0, dim = row * col;

  for (i = 0; i < row; i++)
    for (j = 0; j < col; j++)
      norm2 += pow(x(i, j), 2);

  if (standardize)
    return norm2 / dim;
  else
    return norm2;
}

MatrixDouble inv(const MatrixDouble &x)
// computes inverse of a matrix based on LU decompositon
{
  int col = x.getCol(), i, j;
  int *indx;
  double **xmat, *colNR, d;
  MatrixDouble inverse(col, col);

  assert(col == x.getRow());

  xmat = dmatrix(1, col, 1, col);
  colNR = dvector(1, col);
  indx = ivector(1, col);
  for (i = 0; i < col; i++)
    for (j = 0; j < col; j++)
      xmat[i + 1][j + 1] = x[i][j];

  ludcmp(xmat, col, indx, &d);

  for (j = 1; j <= col; j++) {
    for (i = 1; i <= col; i++)
      colNR[i] = 0.0;
    colNR[j] = 1.0;
    lubksb(xmat, col, indx, colNR);
    for (i = 1; i <= col; i++)
      inverse[i - 1][j - 1] = colNR[i];
  }

  free_ivector(indx, 1, col);
  free_dvector(colNR, 1, col);
  free_dmatrix(xmat, 1, col, 1, col);

  return inverse;
}

MatrixDouble AinvB(const MatrixDouble &A, const MatrixDouble &B)
// efficient way to compute A-inverse times B
{
  int col = A.getCol(), i, j;
  int *indx;
  double **xmat, *colNR, d;
  //const double tol = 1.0e-16;
  MatrixDouble product(col, B.getCol());

  assert(col == A.getRow());
  assert(col == B.getRow());

  xmat = dmatrix(1, col, 1, col);
  colNR = dvector(1, B.getCol());
  indx = ivector(1, col);
  for (i = 0; i < col; i++)
    for (j = 0; j < col; j++)
      xmat[i + 1][j + 1] = A[i][j];

  ludcmp(xmat, col, indx, &d);

  for (j = 1; j <= B.getCol(); j++) {
    for (i = 1; i <= col; i++)
      colNR[i] = B[i - 1][j - 1];   // col of B instead of unit vector
    lubksb(xmat, col, indx, colNR);
    for (i = 1; i <= col; i++)
      product[i - 1][j - 1] = colNR[i];
  }

  free_ivector(indx, 1, col);
  free_dvector(colNR, 1, B.getCol());
  free_dmatrix(xmat, 1, col, 1, col);

  return product;
}

void choldc(double **a, int n, double p[])
{
	void nrerror(char error_text[]);
	int i,j,k;
	float sum;

	for (i=1;i<=n;i++) {
		for (j=i;j<=n;j++) {
			for (sum=a[i][j],k=i-1;k>=1;k--) sum -= a[i][k]*a[j][k];
			if (i == j) {
        assert (sum > 0);
				if (sum <= 0.0)
					nrerror("choldc failed");
				p[i]=sqrt(sum);
			} else a[j][i]=sum/p[i];
		}
	}
}

MatrixDouble rootChol(const MatrixDouble &mat)
// computes square root of matrix using Cholesky decomposition (lower triang.)
{
  int col = mat.getCol(), row = mat.getRow(), i, j;
  assert (col == row);
  double **a, *p;
  MatrixDouble matRoot = nullMatrixDouble(row, col);

  a = dmatrix(1, row, 1, row);
  p = dvector(1, row);

  for (i = 0; i < row; i ++)
    for (j = 0; j < row; j++)
      a[i + 1][j + 1] = mat(i, j);

  choldc(a, col, p);

  for (i = 0; i < row; i ++) {
    matRoot(i, i) = p[i + 1];
    for (j = 0; j < i; j++) {
      matRoot(i, j) = a[i + 1][j + 1];
    }
  }

  free_dmatrix(a, 1, row, 1, row);
  free_dvector(p, 1, row);

  return matRoot;
}

MatrixDouble invSVD(const MatrixDouble &mat)
// calculates inverse of a square matrix using SVD of Numerical Recipes
{
  int col = mat.getCol(), row = mat.getRow(), i, j, k;
  assert (col == row);
  double eps = 1.0e-6;

  double **a, *w, **v;
  MatrixDouble matInv(row, col);
  VectorDouble wVec(col);

  a = dmatrix(1, row, 1, row);
  v = dmatrix(1, row, 1, row);
  w = dvector(1, row);

  for (i = 0; i < row; i ++)
    for (j = 0; j < row; j++)
      a[i + 1][j + 1] = mat(i, j);

  svdcmp(a, col, col, w, v);

  for (j = 1; j <= row; j++) {
    wVec[j - 1] = w[j];
  }

  assert(min(wVec) > eps);

  for (i = 0; i < row; i ++)
    for (j = 0; j < row; j++) {
      matInv(i, j) = 0;
      for (k = 0; k < row; k++)
      matInv(i, j) += v[i + 1][k + 1] * a[j + 1][k + 1] / w[k + 1];
    }

  free_dmatrix(a, 1, row, 1, row);
  free_dmatrix(v, 1, row, 1, row);
  free_dvector(w, 1, row);

  return matInv;
}

MatrixDouble powSVD(const MatrixDouble &mat, double power)
// raises matrix to a power using SVD of Numerical Recipes
{
  int col = mat.getCol(), row = mat.getRow(), i, j, k;
  double eps = 1.0e-6;
  assert (col == row);

  double **a, *w, **v;
  MatrixDouble matInv(row, col);
  VectorDouble wVec(col);

  a = dmatrix(1, row, 1, row);
  v = dmatrix(1, row, 1, row);
  w = dvector(1, row);

  for (i = 0; i < row; i ++)
    for (j = 0; j < row; j++)
      a[i + 1][j + 1] = mat(i, j);

  svdcmp(a, col, col, w, v);

  for (j = 1; j <= row; j++) {
    wVec[j - 1] = w[j];
    if (fabs(w[j]) < eps)
      w[j] = 0;
  }

  if (power < 0) {
    assert(min(wVec) > eps);
  }

  for (i = 0; i < row; i ++)
    for (j = 0; j < row; j++) {
      matInv(i, j) = 0;
      for (k = 0; k < row; k++)
      matInv(i, j) += v[i + 1][k + 1] * a[j + 1][k + 1] * pow(w[k + 1], power);
    }

  free_dmatrix(a, 1, row, 1, row);
  free_dmatrix(v, 1, row, 1, row);
  free_dvector(w, 1, row);

  return matInv;
}

MatrixDouble powSVD1(const MatrixDouble &mat, double power)
// raises matrix to a power using SVD of Numerical Recipes
// "cheat by making all eigenvalues positive"
{
  int col = mat.getCol(), row = mat.getRow(), i, j, k;
  double eps = 1.0e-6;
  assert (col == row);

  double **a, *w, **v;
  MatrixDouble matInv(row, col);
  VectorDouble wVec(col);

  a = dmatrix(1, row, 1, row);
  v = dmatrix(1, row, 1, row);
  w = dvector(1, row);

  for (i = 0; i < row; i ++)
    for (j = 0; j < row; j++)
      a[i + 1][j + 1] = mat(i, j);

  svdcmp(a, col, col, w, v);

  for (j = 1; j <= row; j++) {
    w[j] = fabs(w[j]);
    if (fabs(w[j]) < eps)
      w[j] = eps;
  }

  for (i = 0; i < row; i ++)
    for (j = 0; j < row; j++) {
      matInv(i, j) = 0;
      for (k = 0; k < row; k++)
      matInv(i, j) += v[i + 1][k + 1] * a[j + 1][k + 1] * pow(w[k + 1], power);
    }

  free_dmatrix(a, 1, row, 1, row);
  free_dmatrix(v, 1, row, 1, row);
  free_dvector(w, 1, row);

  return matInv;
}

VectorDouble eigenValues(const MatrixDouble &mat)
// calculates eigen values of a square matrix using SVD of Numerical Recipes
{
  int col = mat.getCol(), row = mat.getRow(), i, j;
  VectorDouble values(col);

  assert (col == row);

  double **a, *w, **v;
  MatrixDouble matInv(row, col);

  a = dmatrix(1, row, 1, row);
  v = dmatrix(1, row, 1, row);
  w = dvector(1, row);

  for (i = 0; i < row; i ++)
    for (j = 0; j < row; j++)
      a[i + 1][j + 1] = mat(i, j);

  svdcmp(a, col, col, w, v);

  for (i = 0; i < row; i ++)
    values[i] = w[i + 1];

  free_dmatrix(a, 1, row, 1, row);
  free_dmatrix(v, 1, row, 1, row);
  free_dvector(w, 1, row);

  return values;
}


double det(const MatrixDouble &mat, int takeLogs)
// computes det of a square matrix using SVD of Numerical Recipes
// Idea: multiply the eigenvalues ...
{
  int col = mat.getCol(), row = mat.getRow(), i, j;
  VectorDouble values(col);

  assert (col == row);

  double **a, *w, **v;
  MatrixDouble matInv(row, col);

  a = dmatrix(1, row, 1, row);
  v = dmatrix(1, row, 1, row);
  w = dvector(1, row);

  for (i = 0; i < row; i ++)
    for (j = 0; j < row; j++)
      a[i + 1][j + 1] = mat(i, j);

  svdcmp(a, col, col, w, v);

  for (i = 0; i < row; i ++)
    values[i] = w[i + 1];

  free_dmatrix(a, 1, row, 1, row);
  free_dmatrix(v, 1, row, 1, row);
  free_dvector(w, 1, row);

  if (takeLogs)
    return exp(sum(log(values)));
  else
    return prod(values);
}

