#ifndef _math_h
#define _math_h

#include "vector.h"
#include "matrix.h"

template<class T>
T min(T a, T b)
{
  if (a < b)
    return a;
  else
    return b;
}

template<class T>
T max(T a, T b)
{
  if (a > b)
    return a;
  else
    return b;
}


int min(int a, int b);
int max(int a, int b);
double min(double a, double b);
double max(double a, double b);
int min(const VectorInt &x);
int max(const VectorInt &x);
double min(const VectorDouble &x);
double max(const VectorDouble &x);

VectorInt seq(int from, int to, int by = 1);
VectorDouble seq(double from, double to, double by = 1.0);

int contained (int x, const VectorInt &vec);

VectorDouble fabs(const VectorDouble &x);
VectorDouble pow(const VectorDouble &x, double power);
VectorDouble powProb(const VectorDouble &x, double power);
VectorDouble log(const VectorDouble &x);
VectorDouble exp(const VectorDouble &x);

MatrixDouble center(const MatrixDouble &x, int whichWay);

MatrixDouble outerProduct(const VectorDouble &x);
MatrixDouble outerProduct(const VectorDouble &x, const VectorDouble &y);
double euclidNorm(const VectorDouble &x, int standardize = 0);
double euclidNormSquare(const VectorDouble &x, int standardize = 0);
double supNorm(const VectorDouble &x);

VectorDouble Delta(const VectorDouble &x, int times = 1);
MatrixDouble Delta(const MatrixDouble &x, int rowWise, int times = 1);

MatrixDouble eigen2(const MatrixDouble &x);

double trace(const MatrixDouble &x);
//double detLU(const MatrixDouble &x, int takeLogs = 0);
double frobNorm(const MatrixDouble &x, int standardize = 0);
double frobNormSquare(const MatrixDouble &x, int standardize = 0);
MatrixDouble inv(const MatrixDouble &x);
MatrixDouble AinvB(const MatrixDouble &A, const MatrixDouble &B);
MatrixDouble rootChol(const MatrixDouble &mat);
MatrixDouble invSVD(const MatrixDouble &mat);
MatrixDouble powSVD(const MatrixDouble &mat, double power);
MatrixDouble powSVD1(const MatrixDouble &mat, double power);
VectorDouble eigenValues(const MatrixDouble &mat);
double det(const MatrixDouble &mat, int takeLogs = 0);

#endif // math.h