/*****

File: sort.c
------------
Implementation for the sort.h interface

*****/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "sort.h"
#include "vector.h"
#include "nrutil.h"

void sortt(double *ra, int n)
{
	int l,j,ir,i;
	double rra;

	l=(n >> 1)+1;
	ir=n;
	for (;;) {
		if (l > 1)
			rra=ra[--l];
		else {
			rra=ra[ir];
			ra[ir]=ra[1];
			if (--ir == 1) {
				ra[1]=rra;
				return;
			}
		}
		i=l;
		j=l << 1;
		while (j <= ir) {
			if (j < ir && ra[j] < ra[j+1]) ++j;
			if (rra < ra[j]) {
				ra[i]=ra[j];
				j += (i=j);
			}
			else j=ir+1;
		}
		ra[i]=rra;
	}
}

void sort(double *ra, int n)
{
  ra--;
  sortt(ra, n);
  ra++;
}


void sortt_int(int *ra, int n)
{
	int l,j,ir,i;
	int rra;

	l=(n >> 1)+1;
	ir=n;
	for (;;) {
		if (l > 1)
			rra=ra[--l];
		else {
			rra=ra[ir];
			ra[ir]=ra[1];
			if (--ir == 1) {
				ra[1]=rra;
				return;
			}
		}
		i=l;
		j=l << 1;
		while (j <= ir) {
			if (j < ir && ra[j] < ra[j+1]) ++j;
			if (rra < ra[j]) {
				ra[i]=ra[j];
				j += (i=j);
			}
			else j=ir+1;
		}
		ra[i]=rra;
	}
}

void sort_int(int *ra, int n)
{
  ra--;
  sortt_int(ra, n);
  ra++;
}


void sortt_two(double *ra1, double *ra2, int n)
{
	int l,j,ir,i;
	double rra1, rra2;

	l=(n >> 1)+1;
	ir=n;
	for (;;) {
		if (l > 1) {
			rra1=ra1[--l];
			rra2=ra2[l];
		      }
		else {
			rra1=ra1[ir];
			ra1[ir]=ra1[1];
			rra2=ra2[ir];
			ra2[ir]=ra2[1];
			if (--ir == 1) {
				ra1[1]=rra1;
				ra2[1]=rra2;
				return;
			}
		}
		i=l;
		j=l << 1;
		while (j <= ir) {

			if (j < ir && ra1[j] < ra1[j+1]) ++j;
			if (rra1 < ra1[j]) {
				ra1[i]=ra1[j];
				ra2[i]=ra2[j];
				j += (i=j);
			}
			else j=ir+1;
		}
		ra1[i]=rra1;
		ra2[i]=rra2;
	}
}

void sort_two(double *ra1, double *ra2, int n)
{
  ra1--;
  ra2--;
  sortt_two (ra1, ra2, n);
  ra1++;
  ra2++;
}

VectorDouble sort(const VectorDouble &x)
{
  double *ra;
  int i, n;
  VectorDouble sortedX(x);

  n = x.getSize();
  ra = dvector(0, n - 1);
  for (i = 0; i < n; i++) ra[i] = x[i];
  sort(ra, n);
  for (i = 0; i < n; i++) sortedX[i] = ra[i];
  free_dvector(ra, 0, n - 1);
  return(sortedX);
}

VectorInt sort(const VectorInt &x)
{
  int *ra;
  int i, n;
  VectorInt sortedX(x);

  n = x.getSize();
  ra = ivector(0, n - 1);
  for (i = 0; i < n; i++) ra[i] = x[i];
  sort_int(ra, n);
  for (i = 0; i < n; i++) sortedX[i] = ra[i];
  free_ivector(ra, 0, n - 1);
  return(sortedX);
}

VectorInt order(const VectorDouble &x)
// returns order of x, that is, sort(x) = permute(x, order(x))
{
  int i, n = x.getSize();
  double *ra1, *ra2;
  VectorInt order(n);

  ra1 = dvector(0, n - 1);
  ra2 = dvector(0, n - 1);
  for (i = 0; i < n; i++) {
    ra1[i] = x[i];
    ra2[i] = (double) i;
  }
  sort_two(ra1, ra2, n);
  for (i = 0; i < n; i++)
    order[i] = (int) (ra2[i] / 1);
  free_dvector(ra1, 0, n - 1);
  free_dvector(ra2, 0, n - 1);

  return order;
}

