#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <assert.h>

#include "statfcts.h" 
#include "nrutil.h"
#include "sort.h"
#include "vector.h"
#include "matrix.h"
#include "cdflib.h"
#include "math.h"
#include "tisu.h"

#define D_STYLE "%6.1f"
#define DI_STYLE "%3d"

double unity (const VectorDouble &x)
{
  return(1.0);
}

double sum(double *x, int n)
/* computes arithmetic mean of x[0..n-1] */
{
  double sum;
  int i;
  
  sum = 0;
  for (i = 0; i < n; i++) {
    sum += x[i];
  }
  return (sum);
}

double sum(const VectorDouble &x)
{
  double sum = 0;

  for (int i = 0; i < x.getSize(); i++) {
    sum += x[i];
  }
  return (sum);
}

double prod(const VectorDouble &x)
{
  double product = 1.0;

  for (int i = 0; i < x.getSize(); i++) {
    product *= x[i];
  }
  return (product);
}

int sum(const VectorInt &x)
{
  int sum = 0;

  for (int i = 0; i < x.getSize(); i++) {
    sum += x[i];
  }
  return sum;
}

int prod(const VectorInt &x)
{
  int product = 1;

  for (int i = 0; i < x.getSize(); i++) {
    product *= x[i];
  }
  return product;
}

double mean(double *x, int n)
/* computes arithmetic mean of x[0..n-1] */
{
  double sum;
  int i;

  sum = 0;
  for (i = 0; i < n; i++) {
    sum += x[i];
  }
  return (sum / (double) n);
}

double mean(const VectorDouble &x)
{
  double sum = 0;

  for (int i = 0; i < x.getSize(); i++) {
    sum += x[i];
  }
  return (sum/(double) x.getSize());
}


double square_mean(double *x, int n)
/* computes arithmetic mean of x[0..n-1]^2 */
{
  return(pow(mean(x,n), 2));
}

double meanSquare(const VectorDouble &x)
{
  double sum;

  sum = 0;
  for (int i = 0; i < x.getSize(); i++) {
    sum += pow(x[i], 2);
  }
  return (sum/(double) x.getSize());
}


double median(double *x, int n)
{
  int k;

  sort(x, n);

  if (0 == n % 2) {
    k = n/2;
    return((x[k-1] + x[k]) / 2);
  }
  else {
    k = n/2;
    return (x[k]);
  }
}

double median(const VectorDouble &x)
{
  int n, k;

  n = x.getSize();
  VectorDouble sortX(sort(x));

  if (0 == n % 2) {
    k = n/2;
    return((sortX[k-1] + sortX[k]) / 2);
  }
  else {
    k = n/2;
    return (sortX[k]);
  }
}

double var(double *x, int n)
{
  double x_bar, sum;
  int i;

  sum = 0;
  x_bar = mean(x,n);
  for (i = 0; i < n; i++){
    sum += pow(((x[i] - x_bar)),2);
  }
  return (sum / (double) (n-1));
}

double var(const VectorDouble &x)
{
  double x_bar, sum;
  int i;

  sum = 0;
  x_bar = mean(x);
  for (i = 0; i < x.getSize(); i++){
    sum += pow(((x[i] - x_bar)),2);
  }
  return (sum / (double) (x.getSize()-1));
}

MatrixDouble var(const MatrixDouble &x, int byCol)
//returns Var-Cov matrix either column- or row-wise
{
  int col = x.getCol(), row = x.getRow(), i, j;
  MatrixDouble Var((1- byCol) * row + byCol * col, (1- byCol) * row + byCol * col);

  if (byCol) {
    for (i = 0; i < col; i++)
      for (j = 0; j <= i; j++) {
        Var(i, j) = innerProduct(extractCol(x, i) - mean(extractCol(x, i)),
                                 extractCol(x, j) - mean(extractCol(x, j))) /
                    ((double) row - 1);
        Var(j, i) = Var(i, j);
      }
  }
  else {
    for (i = 0; i < row; i++)
      for (j = 0; j <= i; j++) {
        Var(i, j) = innerProduct(extractRow(x, i) - mean(extractRow(x, i)),
                                 extractRow(x, j) - mean(extractRow(x, j))) /
                    ((double) col - 1);
        Var(j, i) = Var(i, j);
      }
  }

  return Var;
}

MatrixDouble cor(const MatrixDouble &x, int byCol)
//returns correlation matrix either column- or row-wise
{
  int i, j;
  MatrixDouble Var = var(x, byCol), Cor(Var.getRow(), Var.getCol());

  for (i = 0; i < Cor.getRow(); i++)
    Cor(i, i) = 1;
  for (i = 0; i < Cor.getRow(); i++)
    for (j = 0; j < i; j++) {
      Cor(i, j) = Var(i, j) / sqrt(Var(i, i) * Var(j, j));
      Cor(j, i) = Cor(i, j);
    }

  return Cor;
}


double sd(double *x, int n)
{
  return(sqrt(var(x, n)));
}

double sd(const VectorDouble &x)
{
  return(sqrt(var(x)));
}

VectorDouble CLT(const VectorDouble &x, double alpha)
// returns standard CLT confidence interval for the mean
{
  double dist, p, q, z;
  VectorDouble CI(2);

  p = 1 - alpha/2;
  q = 1 - p;
  z = dinvnr(&p, &q);
  dist = z * sd(x) / sqrt(x.getSize());

  CI = CI + mean(x);
  CI[0] -= dist;
  CI[1] += dist;

  return(CI);
}

double norm_euclid(double *x, int n)
{
  double norm;
  int    i;

  norm = 0;
  for (i = 0; i < n; i++)
    norm += pow(x[i], 2);
  return(sqrt(norm));
}


double norm_sup(double *x, int n)
{
  double *dist;
  int    i;

  dist = dvector(0, n-1);

  for (i = 0; i < n; i++)
    dist[i] = fabs(x[i]);
  sort(dist, n);

  return(dist[n-1]);
}


double self_norm_sum_sd(double *x, int n)
/* computes the self-normalized sum sd in Logan, Rice, and Sheep (1973) */
{
  double sum;
  int i;

  sum = 0;
  for (i = 0; i < n; i++) {
    sum += pow(x[i], 2);
  }
  sum /= (double) (n-1);
  return(sqrt(sum));
}


double conf_stable(double *x, int n)
{
  int    mid, i;
  double res;

  mid = n/2;
  res = 0;
  for (i = 0; i < mid; i++)
    res += x[i] - x[mid];
  for (i = mid; i < n; i++)
    res += x[mid] - x[i];
  res = fabs(res);
  return(res);
}

double variance_ratio_robust(double *x, int n)
  /* uses returns, rather than prices! */
{
double  mu, sig_1, sig_q, q_double, m, xq;
int     i, q,  n_new;

  q = 2;

  n_new = n/q * q;
  mu = mean(x, n_new);
  q_double = (double) q;
  m = q_double * (n_new - q_double + 1.0);
  m *= (1.0 - q_double/((double) n_new));

  sig_1 = var(x, n_new);

  xq = 0;
  for (i = 0; i < q; i++)
    xq += x[i];
  sig_q = pow((xq - q_double * mu), 2);
  for (i = q; i < n_new; i++) {
    xq -= x[i-q];
    xq += x[i];
    sig_q += pow((xq - q_double * mu), 2);
  }
  sig_q /= m;

  return(sig_q/sig_1);
}


double variance_ratio_robust_flex(double *x, int n, int q)
  /* - allows q as input
     - uses returns, rather than prices! */
{
double  mu, sig_1, sig_q, q_double, m, xq;
int     i, n_new;

  n_new = n/q * q;
  mu = mean(x, n_new);
  q_double = (double) q;
  m = q_double * (n_new - q_double + 1.0);
  m *= (1.0 - q_double/((double) n_new));

  sig_1 = var(x, n_new);

  xq = 0;
  for (i = 0; i < q; i++)
    xq += x[i];
  sig_q = pow((xq - q_double * mu), 2);
  for (i = q; i < n_new; i++) {
    xq -= x[i-q];
    xq += x[i];
    sig_q += pow((xq - q_double * mu), 2);
  }
  sig_q /= m;

  return(sig_q/sig_1);
}


double variance_ratio_robust_price(double *x, int n)
/* uses prices rather than returns */
{
double  mu, sig_1, sig_q, q_double, m;
int     i, q,  n_new;

  q = 2;

  n_new = n/q * q;
  mu = (x[n-1] - x[0])/(2.0 * (double) n);
  q_double = (double) q;
  m = q_double * (n * q_double - q_double + 1.0);
  m /= (1.0 - 1.0/((double) n));

  sig_1 = 0;
  for (i = 1; i < n_new; i++) {
    sig_1 += pow((x[i] - x[i-1] - mu), 2);
  }
  sig_1 /= (double) (n_new - 1.0);

  sig_q = 0;
  for (i = q; i < n_new; i++) {
    sig_q += pow((x[i] - x[i-q] - q_double * mu), 2);
  }
  sig_q /= (double) (n_new - 1.0);
  sig_q /= m;

  return(sig_q/sig_1);
}


double delta_hat_j(double *x, int n_new, int j)
/* - computes estimate of asymptotic variance of j-th order correlation
     coefficient as needed for the Lo-MacKinley variance ratio test;
   - uses returns, rather than prices */
{
double  mu, numerator, denominator;
int     i;

  mu = mean(x, n_new);

  numerator = 0;
  for (i = j; i < n_new; i++) {
    numerator += pow((x[i] - mu), 2) * pow((x[i-j] - mu), 2);
  }
  denominator = 0;
  for (i = 0; i < n_new; i++) {
  denominator += pow((x[i] - mu), 2);
  }
  denominator = pow(denominator, 2);
  return(numerator/denominator);
}

double V_hat_q(double *x, int n, int q)
/* - computes asymptotic variance for the variance ratio statistic M-bar(q)
     according to Lo and MacKinley;
   - uses returns rather than prices */
{
double  q_double, result, dummie;
int     j,  n_new;

  n_new = n/q * q;
  q_double = (double) q;
  result = 0;
  for (j=1; j<q; j++) {
    dummie = 2*(q_double - (double) j)/q_double;
    result += pow(dummie, 2) * delta_hat_j(x, n_new, j);
  }
  return(result);
}


double tilt_log(double x)
{
  return(0.2 + log(1.0 + fabs(x)));
}



double fit_abb(double *x, double *y, int ndata)
{
	int i;

	double t,sxoss,sx=0.0,sy=0.0,st2=0.0,ss;
        double b;

	b=0.0;
	for (i=1;i<=ndata;i++) {
	  sx += x[i];
	  sy += y[i];
	}
	ss=ndata;
	sxoss=sx/ss;
	for (i=1;i<=ndata;i++) {
	  t=x[i]-sxoss;
	  st2 += t*t;
	  b += t*y[i];
	}
	
	b /= st2;
/*	a=(sy-sx*(b))/ss;
	siga=sqrt((1.0+sx*sx/(ss*st2))/ss);
	sigb=sqrt(1.0/st2); */
        return(b);
}


/********************
double ar1_fit(double *x, int n)
{
  return(fit_abb(x, x+1, n-1));
}
**********/

double ar1_fit(double *x, int n)
{
  double num, denom;
  int i;

  num = 0;
  denom = 0;
  for (i = 0; i < n - 1; i++) {
    num += x[i] * x[i+1];
    denom += pow(x[i], 2);
  }
  return(num/denom);
}


double ar1_fit_bc(double *x, int n)
{
  double ar1;

  ar1 = fit_abb(x, x+1, n-1);
  return(ar1 + (1.0 + 3 * ar1)/((double) n));
}

/************

double ar1_fit_sd(double *x, int n)
{
  return(sqrt(fabs(1 - pow(ar1_fit(x, n), 2))/((double) (n))));
}
**********/

double ar1_fit_sd(double *x, int n)
{
  double num, denom, rho_hat;
  int i;

  rho_hat = ar1_fit(x, n);
  num = 0;
  denom = 0;
  for (i = 0; i < n - 1; i++) {
    num += pow(x[i+1] - rho_hat * x[i], 2);
    denom += pow(x[i], 2);
  }
  num /= (double) (n - 1);
  return(sqrt(num/denom));
}


double ar1_inter_fit(double *x, int n)
/* - fits AR(1) coefficient when intercept is included in regression
   - uses formula beta.hat = S_{xy}/S_{xx} when y is regressed on x
*/
{
  double num, denom, x_bar, y_bar;
  int i;

  x_bar = mean(x, n - 1);
  y_bar = mean(x + 1, n - 1);
  num = 0;
  denom = 0;
  for (i = 0; i < n - 1; i++) {
    num += (x[i] - x_bar) * (x[i+1] - y_bar);
    denom += pow(x[i] - x_bar, 2);
  }
  return(num/denom);
}

double ar1_inter_fit_sd(double *x, int n)
/* - estimates root(n) * SD of fitted AR(1) coefficient when intercept
     is included in regression
   - uses standard formula
*/
{
  double num, denom, x_bar, y_bar, alpha_hat, beta_hat;
  int i;

  x_bar = mean(x, n - 1);
  y_bar = mean(x + 1, n - 1);
  num = 0;
  denom = 0;                         /* denom will be S_xx */
  for (i = 0; i < n - 1; i++) {
    num += (x[i] - x_bar) * (x[i+1] - y_bar);
    denom += pow(x[i] - x_bar, 2);
  }
  beta_hat = (num/denom);
  alpha_hat = y_bar - beta_hat * x_bar;
  num = 0;
  for (i = 0; i < n - 1; i++) {
    num += pow(x[i+1] - alpha_hat - beta_hat * x[i], 2);
  }
  num /= (double) (n - 3);
  return(sqrt((double) n) * sqrt(num/denom)); /* root(n) * root of s^2/S_xx */
}


VectorDouble linefitSS(const VectorDouble &x, const VectorDouble &y, int intercept)
/*
- does simple linear regression of y on x
- if intercecpt == 1, intercept is included in the fit
- returned are fitted coefficient(s) and sum of squared residuals
*/
{
  int n = x.getSize();
  assert (n == y.getSize());
  if (intercept)
    assert(n > 1);
  else
    assert(n > 0);

  if (intercept) {
    double Sx, Sxx, Sy;
    VectorDouble output(3);
    Sx = sum(x);
    Sy = sum(y);
    Sxx = innerProduct(x - Sx / ((double) n));
    //cout << x << " ";
    output[1] = innerProduct(x - Sx / ((double) n), y) / Sxx;
    output[0] = (Sy - Sx * output[1]) / ((double) n);
    output[2] = innerProduct(y - (output[0] + output[1] * x));
    return output;
  }
  else {
    double Sxx = innerProduct(x, x);
    VectorDouble output(2);
    output[0] = innerProduct(x, y) / Sxx;
    output[1] = innerProduct(y - output[0] * x);
    return output;
  }
}


VectorDouble linefit(const VectorDouble &x, const VectorDouble &y, int intercept)
/*
- does simple linear regression of y on x
- if intercecpt == 1, intercept is included in the fit
- returned are fitted coefficient(s), 'unbiased' estimate of sigma,
  and standard error(s) of coefficients
*/
{
  int n = x.getSize();
  assert (n == y.getSize());

  if (intercept) {
    double Sx, Sxx, Sy;
    VectorDouble output(5);
    Sx = sum(x);
    Sy = sum(y);
    Sxx = innerProduct(x - Sx / ((double) n));
    output[1] = innerProduct(x - Sx / ((double) n), y) / Sxx;
    output[0] = (Sy - Sx * output[1]) / ((double) n);
    output[2] = innerProduct(y - (output[0] + output[1] * x)) / ((double) (n - 2));
    output[2] = pow(output[2], 0.5);
    output[3] = output[2] * pow((1 + pow(Sx, 2) / (((double) n) * Sxx)) / ((double) n), 0.5);
    output[4] = output[2] / pow(Sxx, 0.5);
    return output;
  }
  else {
    double Sxx = innerProduct(x, x);
    VectorDouble output(3);
    output[0] = innerProduct(x, y) / Sxx;
    output[1] = pow(innerProduct(y - output[0] * x) / ((double) (n - 1)), 0.5);
    output[2] = output[1] / pow(Sxx, 0.5);
    return output;
  }
}


void svbksb(double **u, double *w, double **v, int m, int n, double *b, double *x)
{
	int jj,j,i;
	double s,*tmp;

	tmp=dvector(1,n);
	for (j=1;j<=n;j++) {
		s=0.0;
		if (w[j]) {
			for (i=1;i<=m;i++) s += u[i][j]*b[i];
			s /= w[j];
		}
		tmp[j]=s;
	}
	for (j=1;j<=n;j++) {
		s=0.0;
		for (jj=1;jj<=n;jj++) s += v[j][jj]*tmp[jj];
		x[j]=s;
	}
	free_dvector(tmp,1,n);
}



/*static double maxarg1,maxarg2;
#define MAX(a,b) (maxarg1=(a),maxarg2=(b),(maxarg1) > (maxarg2) ?\
	(maxarg1) : (maxarg2)) */
#define SIGN(a,b) ((b) >= 0.0 ? fabs(a) : -fabs(a))


double pythag(double a, double b)
{
        double absa,absb;

	absa=fabs(a);
	absb=fabs(b);
	if (absa > absb) {
	  return (absa*sqrt(1.0+pow((absb/absa),2)));
	}
	else {
	  return (absb == 0.0 ? 0.0 : absb*sqrt(1.0+pow((absa/absb),2)));
	}
}


// temporarily the its check has been disabled !!!!!!!!!

void svdcmp(double **a, int m, int n, double *w, double **v)
/*
- computes SVD A = U % W % V', A is an m x n matrix
- U replaces A on output
- V is returned as V not as V'
- W returns the diagonal elements only
*/
{
	int flag,i,its,j,jj,k,l,nm;
	double c,f,h,s,x,y,z;
	double anorm=0.0,g=0.0,scale=0.0;
	double *rv1;

	if (m < n) nrerror("SVDCMP: You must augment A with extra zero rows");
	rv1=dvector(1,n);
	for (i=1;i<=n;i++) {
		l=i+1;
		rv1[i]=scale*g;
		g=s=scale=0.0;
		if (i <= m) {
			for (k=i;k<=m;k++) scale += fabs(a[k][i]);
			if (scale) {
				for (k=i;k<=m;k++) {
					a[k][i] /= scale;
					s += a[k][i]*a[k][i];
				}
				f=a[i][i];
				g = -SIGN(sqrt(s),f);
				h=f*g-s;
				a[i][i]=f-g;
				if (i != n) {
					for (j=l;j<=n;j++) {
						for (s=0.0,k=i;k<=m;k++) s += a[k][i]*a[k][j];
						f=s/h;
						for (k=i;k<=m;k++) a[k][j] += f*a[k][i];
					}
				}
				for (k=i;k<=m;k++) a[k][i] *= scale;
			}
		      }
		w[i]=scale*g;
		g=s=scale=0.0;
		if (i <= m && i != n) {
			for (k=l;k<=n;k++) scale += fabs(a[i][k]);
			if (scale) {
				for (k=l;k<=n;k++) {
					a[i][k] /= scale;
					s += a[i][k]*a[i][k];
				}
				f=a[i][l];
				g = -SIGN(sqrt(s),f);
				h=f*g-s;
				a[i][l]=f-g;
				for (k=l;k<=n;k++) rv1[k]=a[i][k]/h;
				if (i != m) {
					for (j=l;j<=m;j++) {
						for (s=0.0,k=l;k<=n;k++) s += a[j][k]*a[i][k];
						for (k=l;k<=n;k++) a[j][k] += s*rv1[k];
					}
				}
				for (k=l;k<=n;k++) a[i][k] *= scale;
			}
		}
		anorm=MAX(anorm,(fabs(w[i])+fabs(rv1[i])));
	      } /*end i loop*/
	for (i=n;i>=1;i--) {
		if (i < n) {
			if (g) {
				for (j=l;j<=n;j++)
					v[j][i]=(a[i][j]/a[i][l])/g;
				for (j=l;j<=n;j++) {
					for (s=0.0,k=l;k<=n;k++) s += a[i][k]*v[k][j];
					for (k=l;k<=n;k++) v[k][j] += s*v[k][i];
				}
			}
			for (j=l;j<=n;j++) v[i][j]=v[j][i]=0.0;
		}
		v[i][i]=1.0;
		g=rv1[i];
		l=i;
	}
	for (i=n;i>=1;i--) {
		l=i+1;
		g=w[i];
		if (i < n)
			for (j=l;j<=n;j++) a[i][j]=0.0;
		if (g) {
			g=1.0/g;
			if (i != n) {
				for (j=l;j<=n;j++) {
					for (s=0.0,k=l;k<=m;k++) s += a[k][i]*a[k][j];
					f=(s/a[i][i])*g;
					for (k=i;k<=m;k++) a[k][j] += f*a[k][i];
				}
			}
			for (j=i;j<=m;j++) a[j][i] *= g;
		} else {
			for (j=i;j<=m;j++) a[j][i]=0.0;
		}
		++a[i][i];
	}
	for (k=n;k>=1;k--) {
		for (its=1;its<=100;its++) {
			flag=1;
			for (l=k;l>=1;l--) {
				nm=l-1;
				if (fabs(rv1[l])+anorm == anorm) {
					flag=0;
					break;
				}
				if (fabs(w[nm])+anorm == anorm) break;
			}
			if (flag) {
				c=0.0;
				s=1.0;
				for (i=l;i<=k;i++) {
					f=s*rv1[i];
					if (fabs(f)+anorm != anorm) {
						g=w[i];
						h=pythag(f,g);
						w[i]=h;
						h=1.0/h;
						c=g*h;
						s=(-f*h);
						for (j=1;j<=m;j++) {
							y=a[j][nm];
							z=a[j][i];
							a[j][nm]=y*c+z*s;
							a[j][i]=z*c-y*s;
						}
					}
				}
			}
			z=w[k];
			if (l == k) {
				if (z < 0.0) {
					w[k] = -z;
					for (j=1;j<=n;j++) v[j][k]=(-v[j][k]);
				}
				break;
			}
      //assert(its != 50);
			//if (its == 50) nrerror("No convergence in 50 SVDCMP iterations");
			x=w[l];
			nm=k-1;
			y=w[nm];
			g=rv1[nm];
			h=rv1[k];
			f=((y-z)*(y+z)+(g-h)*(g+h))/(2.0*h*y);
			g=pythag(f,1.0);
			f=((x-z)*(x+z)+h*((y/(f+SIGN(g,f)))-h))/x;
			c=s=1.0;
			for (j=l;j<=nm;j++) {
				i=j+1;
				g=rv1[i];
				y=w[i];
				h=s*g;
				g=c*g;
				z=pythag(f,h);
				rv1[j]=z;
				c=f/z;
				s=h/z;
				f=x*c+g*s;
				g=g*c-x*s;
				h=y*s;
				y=y*c;
				for (jj=1;jj<=n;jj++) {
					x=v[jj][j];
					z=v[jj][i];
					v[jj][j]=x*c+z*s;
					v[jj][i]=z*c-x*s;
				}
				z=pythag(f,h);
				w[j]=z;
				if (z) {
					z=1.0/z;
					c=f*z;
					s=h*z;
				}
				f=(c*g)+(s*y);
				x=(c*y)-(s*g);
				for (jj=1;jj<=m;jj++) {
					y=a[jj][j];
					z=a[jj][i];
					a[jj][j]=y*c+z*s;
					a[jj][i]=z*c-y*s;
				}
			}
			rv1[l]=0.0;
			rv1[k]=f;
			w[k]=x;
		      } /*end its loop*/
	      } /* end k loop */
	free_dvector(rv1,1,n);
}

#define TOL 1.0e-5

void svdfit(double **x, double *y, double *a, int ndata, int ma,
            double **u, double **v, double *w)
{
	int j,i;
	double wmax,thresh,*b;
	//void svdcmp(),svbksb();

	b=dvector(1,ndata);

	for (i=1;i<=ndata;i++) {
	  for (j=1; j<=ma; j++)
	    u[i][j]=x[(i-1)][(j-1)];
	  b[i]=y[i-1];
	}
	svdcmp(u,ndata,ma,w,v);
	wmax=0.1;
	for (j=1;j<=ma;j++)
		if (w[j] > wmax) wmax=w[j];
	thresh=TOL*wmax;
	for (j=1;j<=ma;j++)
		if (w[j] < thresh) w[j]=0.0;
	svbksb(u,w,v,ndata,ma,b,a-1);
	free_dvector(b,1,ndata);
}

void svdfit_compact(double **x, double *y, double *a, int ndata, int ma)
{
	int j,i;
	double wmax,thresh,*b,**u, **v, *w;
	//void svdcmp(),svbksb();

	b=dvector(1,ndata);
  w=dvector(0,ndata);
  u=dmatrix(0,ndata,0,ma);
  v=dmatrix(0,ndata,0,ma);

	for (i=1;i<=ndata;i++) {
	  for (j=1; j<=ma; j++)
	    u[i][j]=x[(i-1)][(j-1)];
	  b[i]=y[i-1];
	}
	svdcmp(u,ndata,ma,w,v);
	wmax=0.1;
	for (j=1;j<=ma;j++)
		if (w[j] > wmax) wmax=w[j];
	thresh=TOL*wmax;
	for (j=1;j<=ma;j++)
		if (w[j] < thresh) w[j]=0.0;
	svbksb(u,w,v,ndata,ma,b,a-1);
	free_dvector(b,1,ndata);
  free_dvector(w,0,ndata);
  free_dmatrix(u,0,ndata,0,ma);
  free_dmatrix(v,0,ndata,0,ma);
}


void svdvar(double **v, int ma, double *w, double **cvm)
{
	int k,j,i;
	double sum,*wti;

	wti=dvector(1,ma);
	for (i=1;i<=ma;i++) {
		wti[i]=0.0;
		if (w[i]) wti[i]=1.0/(w[i]*w[i]);
	}
	for (i=1;i<=ma;i++) {
		for (j=1;j<=i;j++) {
			for (sum=0.0,k=1;k<=ma;k++) sum += v[i][k]*v[j][k]*wti[k];
			cvm[j][i]=cvm[i][j]=sum;
		}
	}
	free_dvector(wti,1,ma);
}


void pause()
{
  printf("Press any key to continue: ");
  getchar();
  printf("\n");
}

void svdfit1(double **x, double *y, double *a, int ndata, int ma)
{
	int j,i;
	double wmax,thresh,*b,**u,**v,*w;
	//void svdcmp(),svbksb(),

        u=dmatrix(1, ndata, 1, ma);
        v=dmatrix(1, ma, 1, ma);
        w=dvector(1, ma);
	b=dvector(1,ndata);

	for (i=1;i<=ndata;i++) {
	  for (j=1; j<=ma; j++)
	    u[i][j]=x[(i-1)][(j-1)];
	  b[i]=y[i-1];
	}
	svdcmp(u,ndata,ma,w,v);
	wmax=0.1;
	for (j=1;j<=ma;j++)
		if (w[j] > wmax) wmax=w[j];
	thresh=TOL*wmax;
	for (j=1;j<=ma;j++)
		if (w[j] < thresh) w[j]=0.0;
	svbksb(u,w,v,ndata,ma,b,a-1);

        free_dmatrix(u, 1, ndata, 1, ma);
        free_dmatrix(v, 1, ma, 1, ma);
        free_dvector(w, 1, ma);
	free_dvector(b,1,ndata);
}

double svdfit_one_beta(double **x, double *y, int ndata, int ma, int beta_index)
/* performs linear regression of y[0..ndata-1] on x[0..ndata-1,0..ma-1],
   and returns the beta_index-nth fitted coefficient
   (intercept corresponds to the 1st coefficient).
*/
{
	int j,i;
	double wmax,thresh,*b,**u,**v,*w,*a,beta;
	//void svdcmp(),svbksb();

        u=dmatrix(1, ndata, 1, ma);
        v=dmatrix(1, ma, 1, ma);
        w=dvector(1, ma);
	b=dvector(1,ndata);
	a=dvector(1,ma);

	for (i=1;i<=ndata;i++) {
	  for (j=1; j<=ma; j++)
	    u[i][j]=x[(i-1)][(j-1)];
	  b[i]=y[i-1];
	}

	svdcmp(u,ndata,ma,w,v);
	wmax=0.0;
	for (j=1;j<=ma;j++)
		if (w[j] > wmax) wmax=w[j];
	thresh=TOL*wmax;
	for (j=1;j<=ma;j++)
		if (w[j] < thresh) w[j]=0.0;
	svbksb(u,w,v,ndata,ma,b,a);
	beta=a[beta_index];

        free_dmatrix(u, 1, ndata, 1, ma);
        free_dmatrix(v, 1, ma, 1, ma);
        free_dvector(w, 1, ma);
	free_dvector(b,1,ndata);
	free_dvector(a,1,ma);
        return(beta);
}

VectorDouble svdfit(const MatrixDouble &xMat, const VectorDouble &yVec)
/* performs linear regression of y on xMat and returns vector of
   fitted regression coefficients
*/
{
  assert(xMat.getRow() == yVec.getSize());
	int j,i,ndata=yVec.getSize(),ma=xMat.getCol();
	double wmax,thresh,*b,**u,**v,*w,*a,**x, *y;
	VectorDouble beta(ma);

  y=dvector(0, ndata-1);
  x=dmatrix(0, ndata-1, 0, ma - 1);
  u=dmatrix(1, ndata, 1, ma);
  v=dmatrix(1, ma, 1, ma);
  w=dvector(1, ma);
	b=dvector(1,ndata);
	a=dvector(1,ma);

  for (i = 0; i < ndata; i++) {
    y[i] = yVec[i];
    for (j = 0; j < ma; j++)
      x[i][j] = xMat(i, j);
  }

	for (i=1;i<=ndata;i++) {
	  for (j=1; j<=ma; j++)
	    u[i][j]=x[(i-1)][(j-1)];
	  b[i]=y[i-1];
	}

	svdcmp(u,ndata,ma,w,v);
	wmax=0.0;
	for (j=1;j<=ma;j++)
		if (w[j] > wmax) wmax=w[j];
	thresh=TOL*wmax;
	for (j=1;j<=ma;j++)
		if (w[j] < thresh) w[j]=0.0;
	svbksb(u,w,v,ndata,ma,b,a);
  for (i = 0; i < ma; i++)
    beta[i] = a[i+1];

  free_dvector(y, 0, ndata-1);
  free_dmatrix(x, 0, ndata-1, 0, ma - 1);
  free_dmatrix(u, 1, ndata, 1, ma);
  free_dmatrix(v, 1, ma, 1, ma);
  free_dvector(w, 1, ma);
	free_dvector(b,1,ndata);
	free_dvector(a,1,ma);

  return beta;
}

MatrixDouble svdfitFull(const MatrixDouble &xMat, const VectorDouble &yVec)
/* performs linear regression of y on xMat and returns matrix that contains:
   - first row: vector of fitted coefficients and OLS estimate of sigma
   - second row: vector of fitted residuals
   - third row: vector of fitted observations
*/
{
  assert(xMat.getRow() == yVec.getSize());
	int j,i,ndata=yVec.getSize(),ma=xMat.getCol();
	double wmax,thresh,*b,**u,**v,*w,*a,**x, *y;
  MatrixDouble output(3, ndata), beta(ma, 1);
	VectorDouble yVecHat(ndata), epsHat(ndata);

  y=dvector(0, ndata-1);
  x=dmatrix(0, ndata-1, 0, ma - 1);
  u=dmatrix(1, ndata, 1, ma);
  v=dmatrix(1, ma, 1, ma);
  w=dvector(1, ma);
	b=dvector(1,ndata);
	a=dvector(1,ma);

  for (i = 0; i < ndata; i++) {
    y[i] = yVec[i];
    for (j = 0; j < ma; j++)
      x[i][j] = xMat(i, j);
  }

	for (i=1;i<=ndata;i++) {
	  for (j=1; j<=ma; j++)
	    u[i][j]=x[(i-1)][(j-1)];
	  b[i]=y[i-1];
	}

	svdcmp(u,ndata,ma,w,v);
	wmax=0.0;
	for (j=1;j<=ma;j++)
		if (w[j] > wmax) wmax=w[j];
	thresh=TOL*wmax;
	for (j=1;j<=ma;j++)
		if (w[j] < thresh) w[j]=0.0;
	svbksb(u,w,v,ndata,ma,b,a);

  // fitted coefficients
  for (i = 0; i < ma; i++)
    output(0, i) = beta(i, 0) = a[i+1];

  // fitted observations and residuals
  yVecHat = extractCol(xMat % beta, 0);
  epsHat = yVec - yVecHat;
  assignRow(output, 1, epsHat);
  assignRow(output, 2, yVecHat);

  // sigma-hat
  output(0, ma) = pow(sum(pow(epsHat, 2.0)) / ((double) (ndata - ma)), 0.5);

  free_dvector(y, 0, ndata-1);
  free_dmatrix(x, 0, ndata-1, 0, ma - 1);
  free_dmatrix(u, 1, ndata, 1, ma);
  free_dmatrix(v, 1, ma, 1, ma);
  free_dvector(w, 1, ma);
	free_dvector(b,1,ndata);
	free_dvector(a,1,ma);

  return output;
}

VectorDouble svdfitSigma(const MatrixDouble &xMat, const VectorDouble &yVec)
/* performs linear regression of y on xMat and returns vector that contains:
   vector of fitted coefficients and OLS estimate of sigma
*/
{
  assert(xMat.getRow() == yVec.getSize());
	int j,i,ndata=yVec.getSize(),ma=xMat.getCol();
	double wmax,thresh,*b,**u,**v,*w,*a,**x, *y;
  MatrixDouble beta(ma, 1);
	VectorDouble yVecHat(ndata), epsHat(ndata), output(ma + 1);

  y=dvector(0, ndata-1);
  x=dmatrix(0, ndata-1, 0, ma - 1);
  u=dmatrix(1, ndata, 1, ma);
  v=dmatrix(1, ma, 1, ma);
  w=dvector(1, ma);
	b=dvector(1,ndata);
	a=dvector(1,ma);

  for (i = 0; i < ndata; i++) {
    y[i] = yVec[i];
    for (j = 0; j < ma; j++)
      x[i][j] = xMat(i, j);
  }

	for (i=1;i<=ndata;i++) {
	  for (j=1; j<=ma; j++)
	    u[i][j]=x[(i-1)][(j-1)];
	  b[i]=y[i-1];
	}

	svdcmp(u,ndata,ma,w,v);
	wmax=0.0;
	for (j=1;j<=ma;j++)
		if (w[j] > wmax) wmax=w[j];
	thresh=TOL*wmax;
	for (j=1;j<=ma;j++)
		if (w[j] < thresh) w[j]=0.0;
	svbksb(u,w,v,ndata,ma,b,a);

  // fitted coefficients
  for (i = 0; i < ma; i++)
    output[i] = beta(i, 0) = a[i+1];

  // fitted observations and residuals
  yVecHat = extractCol(xMat % beta, 0);
  epsHat = yVec - yVecHat;
  // sigma-hat
  output[ma] = pow(sum(pow(epsHat, 2.0)) / ((double) (ndata - ma)), 0.5);

  free_dvector(y, 0, ndata-1);
  free_dmatrix(x, 0, ndata-1, 0, ma - 1);
  free_dmatrix(u, 1, ndata, 1, ma);
  free_dmatrix(v, 1, ma, 1, ma);
  free_dvector(w, 1, ma);
	free_dvector(b,1,ndata);
	free_dvector(a,1,ma);

  return output;
}

void svdfit_sd_one_beta(double **x, double *y, int ndata, int ma,
                        int beta_index, double *res)
/*
  - performs linear regression of y[0..ndata-1] on x[0..ndata-1,0..ma-1],
   and returns the beta_index-nth fitted coefficient
   (intercept corresponds to the 1st coefficient).
  - fitted coefficient will be put in res[0] and its standard error in res[1]

*/
{
	int k,j,i;
	double wmax,thresh,*b,**u,**v,*w,*a,*wti,sum,s2,yhat;

        u=dmatrix(1, ndata, 1, ma);
        v=dmatrix(1, ma, 1, ma);
        w=dvector(1, ma);
	b=dvector(1,ndata);
	a=dvector(1,ma);
  wti=dvector(1,ma);

	for (i=1;i<=ndata;i++) {
	  for (j=1; j<=ma; j++)
	    u[i][j]=x[(i-1)][(j-1)];
	  b[i]=y[i-1];
	}

	svdcmp(u,ndata,ma,w,v);
	wmax=0.0;
	for (j=1;j<=ma;j++)
		if (w[j] > wmax) wmax=w[j];
	thresh=TOL*wmax;
	for (j=1;j<=ma;j++)
		if (w[j] < thresh) w[j]=0.0;
	svbksb(u,w,v,ndata,ma,b,a);
	res[0]=a[beta_index];

  // compute SE, using modified sdvar
	for (i=1;i<=ma;i++) {
		wti[i]=0.0;
		if (w[i]) wti[i]=1.0/(w[i]*w[i]);
	}
           // compute the approprite entry of (X'X)-inverse
  for (sum=0.0,k=1;k<=ma;k++) sum += v[beta_index][k]*v[beta_index][k]*wti[k];
           // compute s2 (unbiased estimate of sigma-squared)
  for (s2=0.0,i=0;i<ndata;i++) {
    yhat=0;
    for(j=0;j<ma;j++)
      yhat+=x[i][j]*a[j+1];
    s2+=pow((y[i]-yhat),2);
  }
  s2/=(double) (ndata - ma);
  //printf("%f\n", pow(s2, 0.5));

          // put the two together
  res[1]=sqrt(s2*sum);

	free_dvector(wti,1,ma);

        free_dmatrix(u, 1, ndata, 1, ma);
        free_dmatrix(v, 1, ma, 1, ma);
        free_dvector(w, 1, ma);
	free_dvector(b,1,ndata);
	free_dvector(a,1,ma);

}

MatrixDouble VAR1(const MatrixDouble &x, int addInt)
// fits VAR(1) model by OLS
// i-th row of the returned matrix are the regression coefficients of the
// regression of the i-th variable on all the variables (plus possible intercept)
{
  int n = x.getRow(), p = x.getCol(), i;
  MatrixDouble A(p, p + addInt), xMat;

  if (addInt)
    xMat = cbind(1, extractRows(x, 0, n - 2));
  else
    xMat = extractRows(x, 0, n - 2);

  for (i = 0; i < p; i++)
    assignRow(A, i, svdfit(xMat, extractCol(x, i)(1, n - 1)));

  return A;
}

MatrixDouble VAR1_SVD(const MatrixDouble &x)
// fits VAR(1) model by OLS and makes SVD adjustment of
// Andrews and Mohanan (1992)
{
  int n = x.getRow(), p = x.getCol(), i, j, k;
  MatrixDouble A(p, p), xMat;

  xMat = extractRows(x, 0, n - 2);

  for (i = 0; i < p; i++)
    assignRow(A, i, svdfit(xMat, extractCol(x, i)(1, n - 1)));

  double **a, *w, **v;

  a = dmatrix(1, p, 1, p);
  v = dmatrix(1, p, 1, p);
  w = dvector(1, p);

  for (i = 0; i < p; i ++)
    for (j = 0; j < p; j++)
      a[i + 1][j + 1] = A(i, j);

  svdcmp(a, p, p, w, v);

  for (i = 1; i <= p; i++) {
    if (w[i] > 0.97) w[i] = 0.97;
    if (w[i] < -0.97) w[i] = -0.97;
  }

  for (i = 0; i < p; i ++)
    for (j = 0; j < p; j++) {
      A(i, j) = 0;
      for (k = 0; k < p; k++)
      A(i, j) += a[i + 1][k + 1] * v[j + 1][k + 1] * w[k + 1];
    }

  free_dmatrix(a, 1, p, 1, p);
  free_dmatrix(v, 1, p, 1, p);
  free_dvector(w, 1, p);

  return A;

}

MatrixDouble VAR1_Resid(const MatrixDouble &x, const MatrixDouble &A)
// computes residuals of VAR(1) fit
// automatically checks whether intercept was used in the fitting process
{
  assert(x.getCol() == A.getRow());
  assert(x.getCol() <= A.getCol() && A.getCol() <= x.getCol() + 1);
  int n = x.getRow(), p = x.getCol(), i, j, addInt = 0;
  MatrixDouble res(n - 1, p);

  if (A.getCol() == x.getCol() + 1)
    addInt = 1;

  for (i = 0; i < n - 1; i++) {
    for (j = 0; j < p; j++) {
      if (addInt)
        res(i, j) = x(i + 1, j) -
                    innerProduct(extractRow(A, j), c(1, (extractRow(x, i))));
      else
        res(i, j) = x(i + 1, j) -
                    innerProduct(extractRow(A, j), (extractRow(x, i)));
    }
  }

  return res;
}


double VAR1_SVD(const VectorDouble &x)
// fits VAR(1) model by OLS and makes SVD adjustment of
// Andrews and Mohanan (1992)
{
  int n = x.getSize();
  double a;

  a = innerProduct(x(0, n - 2), x(1, n - 1)) / innerProduct(x(0, n - 2), x(0, n - 2));
  if (a > 0.97)
    return 0.97;
  if (a < - 0.97)
    return -0.97;
  return a;
}


#define NRANSI
#define TINY 1.0e-20;

void ludcmp(double **a, int n, int *indx, double *d)
{
	int i,imax,j,k;
	double big,dum,sum,temp;
	double *vv;

	vv=dvector(1,n);
	*d=1.0;
	for (i=1;i<=n;i++) {
		big=0.0;
		for (j=1;j<=n;j++)
			if ((temp=fabs(a[i][j])) > big) big=temp;
		if (big == 0.0) nrerror("Singular matrix in routine ludcmp");
		vv[i]=1.0/big;
	}
	for (j=1;j<=n;j++) {
		for (i=1;i<j;i++) {
			sum=a[i][j];
			for (k=1;k<i;k++) sum -= a[i][k]*a[k][j];
			a[i][j]=sum;
		}
		big=0.0;
		for (i=j;i<=n;i++) {
			sum=a[i][j];
			for (k=1;k<j;k++)
				sum -= a[i][k]*a[k][j];
			a[i][j]=sum;
			if ( (dum=vv[i]*fabs(sum)) >= big) {
				big=dum;
				imax=i;
			}
		}
		if (j != imax) {
			for (k=1;k<=n;k++) {
				dum=a[imax][k];
				a[imax][k]=a[j][k];
				a[j][k]=dum;
			}
			*d = -(*d);
			vv[imax]=vv[j];
		}
		indx[j]=imax;
		if (a[j][j] == 0.0) a[j][j]=TINY;
		if (j != n) {
			dum=1.0/(a[j][j]);
			for (i=j+1;i<=n;i++) a[i][j] *= dum;
		}
	}
	free_dvector(vv,1,n);
}
#undef TINY
#undef NRANSI

void lubksb(double **a, int n, int *indx, double b[])
{
	int i,ii=0,ip,j;
	double sum;

	for (i=1;i<=n;i++) {
		ip=indx[i];
		sum=b[ip];
		b[ip]=b[i];
		if (ii)
			for (j=ii;j<=i-1;j++) sum -= a[i][j]*b[j];
		else if (sum) ii=i;
		b[i]=sum;
	}
	for (i=n;i>=1;i--) {
		sum=b[i];
		for (j=i+1;j<=n;j++) sum -= a[i][j]*b[j];
		b[i]=sum/a[i][i];
	}
}


VectorDouble LUfit(const MatrixDouble &x, const VectorDouble &y, int sigma)
/*
- does LS fit of y on x by LU decomposition
- faster than SVD but no standard errors can be computed
- if sigma == 1, also sigma-hat is computed
*/
{
  int n = y.getSize(), col = x.getCol(), i, j, k;
  int *indx;
  double **xTx, *xTy, d, s2;
  VectorDouble output(col + 1);

  assert(x.getRow() == n);

  xTx = dmatrix(1, col, 1, col);
  xTy = dvector(1, col);
  indx = ivector(1, col);

  // set up matrix and vector for normal equations xTx % beta-Hat = xTy
  for (i = 1; i <= col; i++) {
    xTy[i] = 0;
    for (k = 1; k <= n; k++)
      xTy[i] += x(k - 1, i - 1) * y[k - 1];
    for (j = 1; j <= i; j++) {
      xTx[i][j] = 0;
      for (k = 1; k <= n; k++) {
        xTx[i][j] += x(k - 1, i - 1) * x(k - 1, j - 1);
      }
      xTx[j][i] = xTx[i][j];
    }
  }

  ludcmp(xTx, col, indx, &d);
  lubksb(xTx, col, indx, xTy);

  for (i = 0; i < col; i++)
    output[i] = xTy[i + 1];

  if (sigma) {
    s2 = innerProduct(y - x % output(0, col - 1)) / ((double) n - col);
    output[col] = sqrt(s2);
  }

  free_ivector(indx, 1, col);
  free_dvector(xTy, 1, col);
  free_dmatrix(xTx, 1, col, 1, col);

  if (sigma)
    return output;
  else
    return output(0, col - 1);
}


MatrixDouble LUfitFull(const MatrixDouble &x, const VectorDouble &y)
/*
- does LS fit of y on x by LU decomposition
- faster than SVD but no standard errors can be computed
- first row: vector of fitted coefficients and OLS estimate of sigma
- second row: vector of fitted residuals
- third row: vector of fitted observations
*/
{
  int n = y.getSize(), col = x.getCol(), i, j, k;
  int *indx;
  double **xTx, *xTy, d, s2;
  VectorDouble yHat, epsHat;
  MatrixDouble output(3, n), beta(col, 1);

  assert(x.getRow() == n);

  xTx = dmatrix(1, col, 1, col);
  xTy = dvector(1, col);
  indx = ivector(1, col);

  // set up matrix and vector for normal equations xTx % beta-Hat = xTy
  for (i = 1; i <= col; i++) {
    xTy[i] = 0;
    for (k = 1; k <= n; k++)
      xTy[i] += x(k - 1, i - 1) * y[k - 1];
    for (j = 1; j <= i; j++) {
      xTx[i][j] = 0;
      for (k = 1; k <= n; k++) {
        xTx[i][j] += x(k - 1, i - 1) * x(k - 1, j - 1);
      }
      xTx[j][i] = xTx[i][j];
    }
  }

  ludcmp(xTx, col, indx, &d);
  lubksb(xTx, col, indx, xTy);

  for (i = 0; i < col; i++)
    output(0, i) = beta(i, 0) = xTy[i + 1];

  s2 = innerProduct(y - extractCol(x % beta, 0)) / ((double) n - col);
  output(0, col) = sqrt(s2);

  // fitted observations and residuals
  yHat = extractCol(x % beta, 0);
  epsHat = y - yHat;
  assignRow(output, 1, epsHat);
  assignRow(output, 2, yHat);

  free_ivector(indx, 1, col);
  free_dvector(xTy, 1, col);
  free_dmatrix(xTx, 1, col, 1, col);

  return output;
}



#undef TOL
#undef SIGN
#undef MAX




