#include <windows.h>
#include <condefs.h>
#include <conio.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <sys/types.h>
#include <iostream.h>
#include <fstream.h>

#pragma hdrstop

#include "sub.h"
#include "rate.h"
#include "statfcts.h"
#include "nrutil.h"
#include "random.h"
#include "tisu.h"
#include "sort.h"
#include "time_sim.h"
#include "math.h"
#include "vector.h"
#include "matrix.h"
#include "cdflib.h"


MatrixDouble subsampling_conf(double theta, const VectorDouble &x, int b,
                        double alpha, int statistic_choice, int sigma_choice,
                        double tau_pow, int rate_hat, int iid, int bc, int with,
                        int nSub)
/*
   - first row contains some statistics of interest
   - next four rows contain P-values and confidence intervals of types
     one-sided lower, one-sided upper, equal-tailed, and symmetric
*/
{
  int     i, n,length;
  double  tau_n, tau_b, thetahat_n, sigma_n, observed, mean_L_n;
  double  two_sided_low, two_sided_up, quantile, one_sided_low, one_sided_up;
  double  (*statisticPtr)(const VectorDouble &), (*sigmaPtr)(const VectorDouble &);
  MatrixDouble res(5, 3);

  n = x.getSize();

  if (0 == iid)
    length = n - b + 1;
  else
    length = nSub;

  if (rate_hat)
    tau_pow = subsampling_range_rate_hat(x, statistic_choice, iid, with, nSub);

  tau_n = pow((double) n, tau_pow);
  tau_b = pow((double) b, tau_pow);

  switch (statistic_choice) {
  case 1:
    statisticPtr = mean;
    break;
  case 2:
    statisticPtr = meanSquare;
    break;
  case 3:
    statisticPtr = median;
    break;
  default:
    statisticPtr =  mean;
    printf("mean chosen by 'default'\n");
  }

  switch (sigma_choice) {
  case 1:
    sigmaPtr = unity;
    break;
  case 2:
    sigmaPtr = sd;
    break;
  default:
    sigmaPtr =  unity;
    printf("unity chosen by 'default'\n");
  }

  VectorDouble S_hat(length), x_sample(b);

  thetahat_n = (*statisticPtr)(x);
  sigma_n = (*sigmaPtr)(x);
  observed= tau_n * (thetahat_n - theta)/sigma_n;

  /* compute subsample statistics according to independent or
     dependent case
  */

  if (0 == iid) {
    if (1 == sigma_choice) {
      for (i = 0; i < length; i++) {
       	S_hat[i] = tau_b * (((*statisticPtr)(x(i, i + b - 1))) - thetahat_n);
         /* here the normalized subsampling statistics are computed */
      }
    }
    else {
      for (i = 0; i < length; i++) {
	      S_hat[i] = tau_b * (((*statisticPtr)(x(i, i + b - 1))) - thetahat_n) /
                                              ((*sigmaPtr)(x(i, i + b - 1)));
	       /* here the normalized subsampling statistics are computed */
      }
    }
  }
  else {
    if (1 == sigma_choice) {
      for (i = 0; i < length; i++) {
        x_sample = sample(x, b, with);
	      S_hat[i] = tau_b * (((*statisticPtr)(x_sample)) - thetahat_n);
      }
    }
    else {
      for (i = 0; i < length; i++) {
        x_sample = sample(x, b, with);
	      S_hat[i] = tau_b * (((*statisticPtr)(x_sample)) - thetahat_n) /
                                            ((*sigmaPtr)(x_sample));
      }
    }
  }

  S_hat = sort(S_hat);

  if (1 == bc) {
    mean_L_n = mean(S_hat);
    S_hat = S_hat - mean_L_n;
    thetahat_n = thetahat_n - mean_L_n/tau_n;
    observed= tau_n * (thetahat_n - theta)/sigma_n;
  }

  two_sided_low = thetahat_n - sigma_n * emp_quantile(1.0 - alpha/2.0, S_hat)/tau_n;
  two_sided_up = thetahat_n - sigma_n * emp_quantile(alpha/2.0, S_hat)/tau_n;
  one_sided_low = thetahat_n - sigma_n * emp_quantile(1.0 - alpha, S_hat)/tau_n;
  one_sided_up = thetahat_n - sigma_n * emp_quantile(alpha, S_hat)/tau_n;

  res(0, 0) = thetahat_n;
  res(0, 1) = sigma_n;
  res(0, 2) = observed;
  res(1, 0) = lower_Pvalue(observed, S_hat);
  res(1, 1) = one_sided_low;
  res(2, 0) = 1.0 - res(1, 0);
  res(2, 2) = one_sided_up;
  res(3, 0) = res(1, 0) < res(2, 0) ? 2 * res(1, 0) : 2 * res(2, 0);
  res(3, 1) = two_sided_low;
  res(3, 2) = two_sided_up;
  for (i = 0; i < length; i++) S_hat[i] = fabs(S_hat[i]);
    S_hat = sort(S_hat);
  //cout << S_hat << endl;
  res(4, 0) = upper_Pvalue(fabs(observed), S_hat);
  quantile = emp_quantile(1.0 - alpha, S_hat);
  res(4, 1) = thetahat_n  - sigma_n * quantile/tau_n;
  res(4, 2) = thetahat_n + sigma_n * quantile/tau_n;

  return(res);
}

MatrixDouble subsampling_conf_stable(const VectorDouble &x, const VectorInt &b_vec,
                         double alpha, int statistic_choice, int sigma_choice,
                         double tau_pow, int rate_hat, int iid, int bc, int with,
                         int nSub, int ma, int stable_span, double stable_thresh)
/* - computes a number of subsampling intervals for a vector of b values
   - namely, lower, upper, equal-tailed, and symmetric
   - a moving average process smoothes the results (over the b values);
     "ma" specifies the length of the moving average
   - results are stored in output (contains interval endpoints, and
     choosen b sizes)
*/

{
  int    i, n_b = b_vec.getSize();
  double stable_min;
  VectorDouble low_smoo1(n_b), up_smoo1(n_b), low_smoo2(n_b), up_smoo2(n_b),
               low_smoo(n_b), up_smoo(n_b), conf_vec(8 * n_b);
  MatrixDouble res(5, 3), CI(5, 3);

  if (!iid)
    ma = 1;

  VectorDouble stable(n_b - ma + 1 - 2 * stable_span);

  // if rate estimation is called for, do it "once and for all" to save time
  if (rate_hat) {
    tau_pow = subsampling_range_rate_hat(x, statistic_choice, iid, with, nSub);
    rate_hat = 0;
  }

  for (i = 0; i <  n_b; i++){

    /* Compute intervals for individual b's */
    CI = subsampling_conf(0, x, b_vec[i], alpha, statistic_choice, sigma_choice,
                          tau_pow, rate_hat, iid, bc, with, nSub);
    conf_vec[i] = CI(3, 1);            /* equal-tailed end points */
    conf_vec[n_b + i] = CI(3, 2);
    conf_vec[2 * n_b + i] = CI(4, 1);  /* symmetric end points */
    conf_vec[3 * n_b + i] = CI(4, 2);
    conf_vec[4 * n_b + i] = CI(1, 1); /* one-sded low */
    conf_vec[5 * n_b + i] = CI(2, 2); /* one-sided up */
    conf_vec[6 * n_b + 10] = CI(0, 0); /* theta.hat */
    conf_vec[6 * n_b + 11] = CI(0, 1); /* standard error */
  }

  /* smooth endpoints using a running mean */

  for (i = 0; i < n_b - ma + 1; i++) {
    low_smoo1[i] = mean(conf_vec(i, i + ma - 1));
    up_smoo1[i] = mean(conf_vec(n_b + i, n_b + i + ma - 1));
    low_smoo2[i] = mean(conf_vec(2 * n_b + i, 2 * n_b + i + ma - 1));
    up_smoo2[i] = mean(conf_vec(3 * n_b + i, 3 * n_b + i + ma - 1));
    low_smoo[i] = mean(conf_vec(4 * n_b + i, 4 * n_b + i + ma - 1));
    up_smoo[i] = mean(conf_vec(5 * n_b + i, 5 * n_b + i + ma - 1));
  }
  for (i = 0; i < n_b - ma + 1; i++) {
    conf_vec[i] = low_smoo1[i];
    conf_vec[n_b + i] = up_smoo1[i];
    conf_vec[2*n_b + i] = low_smoo2[i];
    conf_vec[3*n_b + i] = up_smoo2[i];
    conf_vec[4*n_b + i] = low_smoo[i];
    conf_vec[5*n_b + i] = up_smoo[i];
  }

  /* look for stable region */

  /* do equal-tailed interval first */

  for (i = 0; i < n_b - ma + 1 - 2 * stable_span; i++)
    stable[i] = pow(var(low_smoo1(i, i + 2 * stable_span)), 0.5);

  /* find min of stable[] */
  stable_min = min(sort(stable));

  /* find lower limit confidence interval from stable region */
  i = 0;
  while(stable[i] > stable_thresh * stable_min) i++;
  res(3, 0) = b_vec[i + stable_span];
  res(3, 1) = low_smoo1[i + stable_span];

  for (i = 0; i < n_b - ma + 1 - 2 * stable_span; i++)
    stable[i] = pow(var(up_smoo1(i, i + 2 * stable_span)), 0.5);

  /* find min of stable[] */
  stable_min = min(sort(stable));

  /* find upper limit confidence interval from stable region */
  i = 0;
  while(stable[i] > stable_thresh * stable_min) i++;
  res(3, 0) += b_vec[i + stable_span];
  res(3, 0) = res(3, 0) / 2.0;
  res(3, 2) = up_smoo1[i + stable_span];

  /* repeat exercise for symmetric interval */

  for (i = 0; i < n_b - ma + 1 - 2 * stable_span; i++)
    stable[i] = pow(var(low_smoo2(i, i + 2 * stable_span)), 0.5) +
      pow(var(up_smoo2(i, i + 2 * stable_span)), 0.5);

  /* find min of stable[] */
  stable_min = min(sort(stable));

  /* find confidence interval from stable region */
  i = 0;
  while(stable[i] > stable_thresh * stable_min) i++;
  res(4, 1) = low_smoo2[i + stable_span];
  res(4, 2) = up_smoo2[i + stable_span];
  res(4, 0) = b_vec[i + stable_span];

  /* repeat exercise for lower one-sided interval */

  for (i = 0; i < n_b - ma + 1 - 2 * stable_span; i++)
    stable[i] = pow(var(low_smoo(i, i + 2 * stable_span)), 0.5);

  /* find min of stable[] */
   stable_min = min(sort(stable));

  /* find confidence interval from stable region */
  i = 0;
  while(stable[i] > stable_thresh * stable_min) i++;
  res(1, 1) = low_smoo[i + stable_span];
  res(1, 0) = b_vec[i + stable_span];

  /* repeat exercise for upper one-sided interval */

  for (i = 0; i < n_b - ma + 1 - 2 * stable_span; i++)
    stable[i] = pow(var(up_smoo(i, i + 2 * stable_span)), 0.5);

  /* find min of stable[] */
  stable_min = min(sort(stable));

  /* find confidence interval from stable region */
  i = 0;
  while(stable[i] > stable_thresh * stable_min) i++;
  res(2, 2) = up_smoo[i + stable_span];
  res(2, 0) = b_vec[i + stable_span];

  return(res);
}

MatrixDouble subsampling_conf_stable_sim(int n, const VectorInt &b_vec,
                         double alpha, int statistic_choice, int sigma_choice,
                         double tau_pow, int rate_hat, int iid, int bc, int with,
                         int nSub, int numSim, int ma, int stable_span,
                         double stable_thresh, int print, int read_file)
/*
- Simulation Study for subsampling univariate parameters
- Output are estimated coverage probs for 4 interval types and mean lengths
  of the two two-sided intervals
- Some normal method may also be included
*/
{
  int i, j;
  double theta = 0;
  VectorDouble x(n), coverage(5), CI, xBig;
  MatrixDouble CIs, results(4, 5), length(3, numSim);

  results(0, 0) = (double) n;
  results(0, 1) = theta;
  results(0, 2) = alpha;

  if (read_file) {
    xBig = nullVectorDouble(numSim * n);
    ifstream inFile("stable.out", ios::in);
    if (!inFile) {
      cerr << "File could not be opened" << endl;
      exit(1);
    }
    for (i = 0; i < numSim * n; i++)
      inFile >> xBig[i];
  }

  for (i = 0; i < numSim; i++) {
    cout << i << " ";
    // generate data
    x = rparetoSym(n, 1.5);

    if (read_file)  {
      //cout << xBig.getSize() << " " << i << " " << (i * n) << " " << ((i + 1) * n - 1) << endl;
      x = xBig(i * n, (i + 1) * n - 1);
    }

    // subsampling intervals
    CIs = subsampling_conf_stable(x, b_vec, alpha, statistic_choice, sigma_choice,
                                tau_pow, rate_hat, iid, bc, with, nSub, ma,
                                stable_span, stable_thresh);
    if (CIs(1, 1) <= theta)
      coverage[0] += 1;
    if (theta <= CIs(2, 2))
      coverage[1] += 1;
    for (j = 2; j < 4; j++) {
      if (CIs(j + 1, 1) <= theta && theta <= CIs(j + 1, 2))
        coverage[j] += 1.0;
      length(j - 2, i)= CIs(j + 1, 2) - CIs(j + 1, 1);
    }

    // whatever kind of other interval (like normal interval)
    // do standard CLT interval

    CI = CLT(x, alpha);
    if (CI[0] <= theta && theta <= CI[1])
      coverage[4] += 1;
    length(2, i) += CI[1] - CI[0];
  }

  assignRow(results, 1, coverage / ((double) numSim));
  for (j = 2; j < 5; j++) {
    results(2, j) = mean(extractCol(length, j - 2));
    results(3, j) = median(extractCol(length, j - 2));
  }

  if (print) {
    ofstream outfile("tail.dat", ios::app);
    outfile << "Parameters: n, b_small, b_big, alpha, stat, sd, tau_pow, rate_hat, iid, nSub, numSim" << endl;
    outfile << n << " " << b_vec[0] << " " << b_vec[b_vec.getSize()-1] << " " <<
               alpha << " " << statistic_choice << " " << sigma_choice << " " <<
               tau_pow << " " << rate_hat << " " << iid << " " << nSub << " " <<
               numSim << endl;
    outfile << extractRows(results, 1, results.getRow() - 1) << endl;
  }

  return(results);
}

VectorDouble subsampling_ar_Pv(double arSum0, const VectorDouble &x, int p, int b,
                                 double tau_pow, int stud, int trend)
/*
   - P-values for sum of AR(p) parameters in trend stationary model
   - p is order of model
   - stud == 1 means use studentized approach (tau_pow = 0/0.5)
   - stud == 0 means use unstudentized approach (tau_pow = estimated)
   - trend == 1 means fit a time trend, trend == 0 means not
   - first two numbers are fitted value and estimated
   - next four numbers are P-values of types
     one-sided lower, one-sided upper, equal-tailed, and symmetric
*/

{
  int     i, j, k, n, length;
  double  *y, **x_matrix, **x_sub_matrix, observed;
  double  tau_n, tau_b, alpha_n, se_n, se_n_alt, alpha_b, se_b, ls_fit[2];
  VectorDouble res(6);

  n = x.getSize();
  length = n - p - b + 1;
  VectorDouble S_hat(length);

  if (stud)
    tau_pow = 0;

  tau_n = pow((double) n - p, tau_pow);
  tau_b = pow((double) b, tau_pow);

  y = dvector(0, n - p - 1);
  x_matrix = dmatrix(0, n - p - 1, 0, p + 1);
  x_sub_matrix = dmatrix(0, b - 1, 0, p + 1);

  // set up regression variables
  // the y vector contains x[p] to x[n-p-1] -- in C indices
  // X[,0] is x[p-1] to x[n-2]
  // then come the differenced values
  // the last two columns are the constant and time
  for (i = 0; i < n - p; i++) {
    y[i] = x[i+p];
    x_matrix[i][0] = x[i+p-1];
    for (j = 1; j < p; j++)
      x_matrix[i][j] = x[i+p-j] - x[i+p-j-1];
    x_matrix[i][p] = 1.0;
    x_matrix[i][p+1] = i + p + 1.0;
  }

  if (0 == trend) { // fit without time trend
    svdfit_sd_one_beta(x_matrix, y, n - p, p + 1, 1, ls_fit);
    alpha_n = ls_fit[0];
    se_n = se_n_alt = ls_fit[1];
  }
  else { // fit with time trend
    svdfit_sd_one_beta(x_matrix, y, n - p, p + 2, 1, ls_fit);
    alpha_n = ls_fit[0];
    se_n = se_n_alt = ls_fit[1];
  }
  if (0 == stud)
    se_n = 1.0;
  observed = tau_n * (alpha_n - arSum0)/se_n;

  for (i = 0; i < length; i++) { // compute subsample statistics
    for (j = 0; j < b; j++)
      for (k = 0; k < p + 2; k++)
        x_sub_matrix[j][k] = x_matrix[i+j][k];
    if (0 == trend) { // fit without time trend
      svdfit_sd_one_beta(x_sub_matrix, y + i, b, p + 1, 1, ls_fit);
      alpha_b = ls_fit[0];
      se_b = ls_fit[1];
    }
    else { // fit with time trend
      svdfit_sd_one_beta(x_sub_matrix, y + i, b, p + 2, 1, ls_fit);
      alpha_b = ls_fit[0];
      se_b = ls_fit[1];
    }

    // compute subsample statistics
    if (0 == stud)
      S_hat[i] = tau_b * (alpha_b - alpha_n);
    else
      S_hat[i] = tau_b * (alpha_b - alpha_n)/se_b;
  }

  S_hat = sort(S_hat);

  res[0] = alpha_n;
  res[1] = se_n_alt;
  res[2] = lower_Pvalue(observed, S_hat);
  res[3] = upper_Pvalue(observed, S_hat);
  res[4] = 2 * MIN(res[2], res[3]);
  res[5] = upper_Pvalue(fabs(observed), sort(fabs(S_hat)));

  free_dvector(y, 0, n - p - 1);
  free_dmatrix(x_matrix, 0, n - p - 1, 0, p + 1);
  free_dmatrix(x_sub_matrix, 0, b-1, 0, p + 1);

  return(res);
}


MatrixDouble subsampling_ar_PV_dimi_sim(int n, VectorInt b_vec, double mu,
          double beta, VectorDouble ar, double ma, double alpha, int p_fit, int stud,
          int trend, int num_sim, VectorDouble (*randgen)(int n), int n_start)
/*****
  - Simulation Study for AR(p) model with time trend
  - estimate rejection probs for unit root test (sims for Dimi's paper)
  - do this for fixed block sizes
  - allow p used in the fitting process to be different from the true p
******/
{
  int i, j, k, p, b_len = b_vec.getSize();
  double tau_pow, ar_sum;
  VectorDouble x, res_i;
  MatrixDouble result(5, max(b_len, 5));

  p = ar.getSize();
  ar_sum = sum(ar);

  //put some general info in first row
  result(0, 0) = (double) n;
  result(0, 1) = ar_sum;
  result(0, 2) = ma;
  result(0, 3) = alpha;
  result(0, 4) = (double) num_sim;

	for(i = 0; i < num_sim; i++) {
      cout << i << " ";

    // generate AR(p) sequence
		x = ArmaTrendSim(n, mu, beta, ar, ma, randgen, n_start);

    if (stud)
      tau_pow = 0;
    else {
      tau_pow = 1;
    }

    for (j = 0; j < b_len; j++) {
      res_i = subsampling_ar_Pv(1.0, x, p_fit, b_vec[j], tau_pow, 0, 0);
      for (k = 0; k < 4; k++)
        if (res_i[2 + k] < alpha)
          result(k + 1, j)++;
    }
  }

  cout << endl;

  for (j = 0; j < b_len; j++)
    for (k = 0; k < 4; k++)
      result(k + 1, j) = result(k + 1, j) / ((double) num_sim);

  return result;
}


MatrixDouble subsampling_ar_conf(const VectorDouble &x, int p, int b, double alpha,
                                 double tau_pow, int stud, int trend)
/*
   - CI for sum of AR(p) parameters in trend stationary model
   - p is order of model
   - stud == 1 means use studentized approach (tau_pow = 0/0.5)
   - stud == 0 means use unstudentized approach (tau_pow = estimated)
   - trend == 1 means fit a time trend, trend == 0 means not
   - first row contains some statistics of interest
   - next four rows contain confidence intervals of types
     one-sided lower, one-sided upper, equal-tailed, and symmetric
*/

{
  int     i, j, k, n, length;
  double  *y, **x_matrix, **x_sub_matrix;
  double  tau_n, tau_b, alpha_n, se_n, se_n_alt, alpha_b, se_b, ls_fit[2],
          two_sided_low, two_sided_up, quantile, one_sided_low, one_sided_up;
  MatrixDouble res(5, 2);

  n = x.getSize();
  length = n - p - b + 1;
  VectorDouble S_hat(length);

  if (stud)
    tau_pow = 0;

  tau_n = pow((double) n - p, tau_pow);
  tau_b = pow((double) b, tau_pow);

  y = dvector(0, n - p - 1);
  x_matrix = dmatrix(0, n - p - 1, 0, p + 1);
  x_sub_matrix = dmatrix(0, b - 1, 0, p + 1);

  // set up regression variables
  // the y vector contains x[p] to x[n-p-1] -- in C indices
  // X[,0] is x[p-1] to x[n-2]
  // then come the differenced values
  // the last two columns are the constant and time
  for (i = 0; i < n - p; i++) {
    y[i] = x[i+p];
    x_matrix[i][0] = x[i+p-1];
    for (j = 1; j < p; j++)
      x_matrix[i][j] = x[i+p-j] - x[i+p-j-1];
    x_matrix[i][p] = 1.0;
    x_matrix[i][p+1] = i + p + 1.0;
  }

  if (0 == trend) { // fit without time trend
    svdfit_sd_one_beta(x_matrix, y, n - p, p + 1, 1, ls_fit);
    alpha_n = ls_fit[0];
    se_n = se_n_alt = ls_fit[1];
  }
  else { // fit with time trend
    svdfit_sd_one_beta(x_matrix, y, n - p, p + 2, 1, ls_fit);
    alpha_n = ls_fit[0];
    se_n = se_n_alt = ls_fit[1];
  }
  if (0 == stud)
    se_n = 1.0;

  for (i = 0; i < length; i++) { // compute subsample statistics
    for (j = 0; j < b; j++)
      for (k = 0; k < p + 2; k++)
        x_sub_matrix[j][k] = x_matrix[i+j][k];
    if (0 == trend) { // fit without time trend
      svdfit_sd_one_beta(x_sub_matrix, y + i, b, p + 1, 1, ls_fit);
      alpha_b = ls_fit[0];
      se_b = ls_fit[1];
    }
    else { // fit with time trend
      svdfit_sd_one_beta(x_sub_matrix, y + i, b, p + 2, 1, ls_fit);
      alpha_b = ls_fit[0];
      se_b = ls_fit[1];
    }

    // compute subsample statistics
    if (0 == stud)
      S_hat[i] = tau_b * (alpha_b - alpha_n);
    else
      S_hat[i] = tau_b * (alpha_b - alpha_n)/se_b;
  }

  S_hat = sort(S_hat);
  one_sided_low = alpha_n  - se_n * emp_quantile(1.0 - alpha, S_hat)/tau_n;
  one_sided_up = alpha_n  - se_n * emp_quantile(alpha, S_hat)/tau_n;
  two_sided_low = alpha_n  - se_n * emp_quantile(1.0 - 0.5*alpha, S_hat)/tau_n;
  two_sided_up = alpha_n  - se_n * emp_quantile(0.5*alpha, S_hat)/tau_n;

  res(0, 0) = alpha_n;
  res(0, 1) = se_n_alt;
  res(1, 0) = one_sided_low;
  res(2, 1) = one_sided_up;
  res(3, 0) = two_sided_low;
  res(3, 1) = two_sided_up;
  quantile = emp_quantile(1.0 - alpha, sort(fabs(S_hat)));
  res(4, 0) = alpha_n - se_n * quantile/tau_n;
  res(4, 1) = alpha_n + se_n * quantile/tau_n;

  free_dvector(y, 0, n - p - 1);
  free_dmatrix(x_matrix, 0, n - p - 1, 0, p + 1);
  free_dmatrix(x_sub_matrix, 0, b-1, 0, p + 1);

  return(res);
}


MatrixDouble subsampling_ar_conf_stable(const VectorDouble &x, int p,
                         const VectorInt b_vec, double alpha,
                         double tau_pow, int stud, int trend, int ma,
                         int stable_span, double stable_thresh)
/*
   - computes a number of subsampling intervals for a vector of b values
   - namely, lower, upper, equal-tailed, and symmetric
   - a moving average process smoothes the results (over the b values);
     "ma" specifies the length of the moving average
   - results are stored in output (contains interval endpoints, and
     choosen b sizes)
*/

{
  double *x_copy, *low_smoo1, *up_smoo1, *low_smoo2, *up_smoo2,
         *low_smoo, *up_smoo, *conf_vec;
  double stable_min;
  int    i, b, n_b;
  MatrixDouble res(5, 3), res_b(5, 2);

  n_b = b_vec.getSize();

  VectorDouble stable(n_b - ma + 1 - 2 * stable_span);

  conf_vec = dvector(0, 8 * n_b);
  x_copy = dvector(0, 19);
  low_smoo1 = dvector(0, n_b - 1);
  up_smoo1 = dvector(0, n_b - 1);
  low_smoo2 = dvector(0, n_b - 1);
  up_smoo2 = dvector(0, n_b - 1);
  low_smoo = dvector(0, n_b - 1);
  up_smoo = dvector(0, n_b - 1);

  for (i = 0; i < n_b; i++){

    /* Compute intervals for individual b's */
    b = b_vec[i];
    res_b = subsampling_ar_conf(x, p, b, alpha, tau_pow, stud, trend);
    conf_vec[i] = res_b(3, 0);            /* equal-tailed end points */
    conf_vec[n_b + i] = res_b(3, 1);
    conf_vec[2 * n_b + i] = res_b(4, 0);  /* symmetric end points */
    conf_vec[3 * n_b + i] = res_b(4, 1);
    conf_vec[4 * n_b + i] = res_b(1, 0); /* one-sded low */
    conf_vec[5 * n_b + i] = res_b(2, 1); /* one-sided up */
    res(0, 0) = res_b(0, 0); /* alpha.hat */
    res(0, 1) = res_b(0, 1); /* standard error */
  }

  /* smooth endpoints using a running mean */

  for (i = 0; i < n_b - ma + 1; i++) {
    low_smoo1[i] = mean(conf_vec + i, ma);
    up_smoo1[i] = mean(conf_vec + n_b + i, ma);
    low_smoo2[i] = mean(conf_vec + 2*n_b + i, ma);
    up_smoo2[i] = mean(conf_vec + 3*n_b + i, ma);
    low_smoo[i] = mean(conf_vec + 4*n_b + i, ma);
    up_smoo[i] = mean(conf_vec + 5*n_b + i, ma);
  }
  for (i = 0; i < n_b - ma + 1; i++) {
    conf_vec[i] = low_smoo1[i];
    conf_vec[n_b + i] = up_smoo1[i];
    conf_vec[2*n_b + i] = low_smoo2[i];
    conf_vec[3*n_b + i] = up_smoo2[i];
    conf_vec[4*n_b + i] = low_smoo[i];
    conf_vec[5*n_b + i] = up_smoo[i];
  }

  /* look for stable region */

  /* do equal-tailed interval first */

  for (i = 0; i < n_b - ma + 1 - 2 * stable_span; i++)
    stable[i] = pow(var(low_smoo1 + i, 2 * stable_span + 1), 0.5);

  /* find min of stable[] */
  stable_min = min(sort(stable));

  /* find lower confidence interval endpoint from stable region */
  i = 0;
  while(stable[i] > stable_thresh * stable_min) i++;
  res(3, 0) = low_smoo1[i + stable_span];
  res(3, 2) = b_vec[i + stable_span];

  for (i = 0; i < n_b - ma + 1 - 2 * stable_span; i++)
    stable[i] = pow(var(up_smoo1 + i, 2 * stable_span + 1), 0.5);

  /* find min of stable[] */
  stable_min = min(sort(stable));

  /* find upper confidence interval endpoint from stable region */
  i = 0;
  while(stable[i] > stable_thresh * stable_min) i++;
  res(3, 1) = up_smoo1[i + stable_span];
  res(3, 2) += b_vec[i + stable_span];
  res(3, 2) /= 2.0;



  /* repeat exercise for symmetric interval */

  for (i = 0; i < n_b - ma + 1 - 2 * stable_span; i++)
    stable[i] = pow(var(low_smoo2 + i, 2 * stable_span + 1), 0.5) +
      pow(var(up_smoo2 + i, 2 * stable_span + 1), 0.5);

  /* find min of stable[] */
  stable_min = min(sort(stable));

  /* find confidence interval from stable region */
  i = 0;
  while(stable[i] > stable_thresh * stable_min) i++;
  res(4, 0) = low_smoo2[i + stable_span];
  res(4, 1) = up_smoo2[i + stable_span];
  res(4, 2) = b_vec[i + stable_span];


  /* repeat exercise for lower one-sided interval */

  for (i = 0; i < n_b - ma + 1 - 2 * stable_span; i++)
    stable[i] = pow(var(low_smoo + i, 2 * stable_span + 1), 0.5);

  /* find min of stable[] */
  stable_min = min(sort(stable));

  /* find confidence interval from stable region */
  i = 0;
  while(stable[i] > stable_thresh * stable_min) i++;
  res(1, 0) = low_smoo[i + stable_span];
  res(1, 2) = b_vec[i + stable_span];


  /* repeat exercise for upper one-sided interval */

  for (i = 0; i < n_b - ma + 1 - 2 * stable_span; i++)
    stable[i] = pow(var(up_smoo + i, 2 * stable_span + 1), 0.5);

  /* find min of stable[] */
  stable_min = min(sort(stable));

  /* find confidence interval from stable region */
  i = 0;
  while(stable[i] > stable_thresh * stable_min) i++;
  res(2, 1) = up_smoo[i + stable_span];
  res(2, 2) = b_vec[i + stable_span];

  free_dvector(conf_vec, 0, 8 * n_b);
  free_dvector(x_copy, 0, 19);
  free_dvector(low_smoo1, 0, n_b - 1);
  free_dvector(up_smoo1, 0, n_b - 1);
  free_dvector(low_smoo2, 0, n_b - 1);
  free_dvector(up_smoo2, 0, n_b - 1);
  free_dvector(low_smoo, 0, n_b - 1);
  free_dvector(up_smoo, 0, n_b - 1);

  return(res);
}


MatrixDouble subsampling_ar_conf_stable_sim(int n, VectorInt b_vec, double mu,
          double beta, VectorDouble ar, double alpha, int stud, int trend,
          int num_sim, VectorDouble (*randgen)(int n), int ma_smoo,
          int stable_span, double stable_thresh, int n_start)
/*****
  - Simulation Study for AR(p) model with time trend
  - estimate coverage probability and mean and median lengths of normal and
    subsampling intervals with nominal level 1 - alpha
******/
{
  int i, j, p;
  double tau_pow, ar_sum, clt_quantile, alpha_hat, se_n, eps = 0.0000001;
  VectorDouble tau_pow_i(num_sim), x;
  MatrixDouble result(4, 5), length(num_sim, 3), res_i;

  p = ar.getSize();
  ar_sum = sum(ar);

  clt_quantile = my_dinvnr(1 - alpha/2.0);

  //put some general info in first row
  result(0, 0) = (double) n;
  result(0, 1) = ar_sum;
  result(0, 2) = alpha;
  result(0, 3) = (double) num_sim;

	for(i = 0; i < num_sim; i++) {

    // generate AR(p) sequence
		x = arTrendSim(n, mu, beta, ar, randgen, n_start);

    if (stud)
      tau_pow = 0;
    else {
      //tau_pow = subsampling_ar_range_rate_hat(innov, n, p, trend, 5, 10, 0.01, 0.25, 0.5);
      if (fabs(ar_sum - 1) < eps) tau_pow = 1;
      else tau_pow = 0.5;
      //printf("%f\n", tau_pow);
    }
    tau_pow_i[i] = tau_pow;

    res_i = subsampling_ar_conf_stable(x, p, b_vec, alpha, tau_pow,
                 stud, trend, ma_smoo, stable_span, stable_thresh);

    // CLT
    alpha_hat = res_i(0, 0);
    se_n = res_i(0, 1);
    if (fabs((ar_sum - alpha_hat)/se_n) <= clt_quantile)
      result(1, 4) += 1;
    length(i, 2) = 2 * clt_quantile * se_n;

    // Subsampling intervals
    if (ar_sum > res_i(1, 0))
      result(1, 0) += 1;
    if (ar_sum < res_i(2, 1))
      result(1, 1) += 1;
    for (j = 3; j < 5; j++) {
      if (res_i(j, 0) < ar_sum && ar_sum < res_i(j, 1))
        result(1, j - 1) += 1;
      length(i, j - 3) = res_i(j, 1) - res_i(j, 0);
    }
	}

  for (j = 0; j < 5; j++)
    result(1, j) = result(1, j) / ((double) num_sim);

  for (j = 2; j < 5; j++) {
    result(2, j) = mean(extractCol(length, j - 2));
    result(3, j) = median(extractCol(length, j - 2));
  }

  result(0, 4) = mean(tau_pow_i);

  return(result);
}

MatrixDouble subsampling_ar_conf_b_cal_boot(const VectorDouble &x, int p,
                         const VectorInt b_vec, double alpha,
                         double tau_pow, int stud, int trend, int B, int b_boot,
                         VectorDouble (*boot) (const VectorDouble &, int, int, int),
                         int nStart)
/*
   - estimates block size calibration function by generating pseudo sequences
     from fitted model (assuming intercept is always included)
   - fitted residuals are bootstrapped for this purpose
   - returns the four subsampling intervals (calibrated individually)
   - results are output (contains interval endpoints, and
     choosen b sizes)
*/
{
  int n = x.getSize(), n_b = b_vec.getSize(), i, j;
  double arSum, eps = 0.00001, min_diff;
  VectorInt b_opt(5);
  VectorDouble xStar(n), ar(p + 2), arStar(p + 2), yVec(n - p), innov(n - p),
               diff(n_b), res_b;
  MatrixDouble h(5, n_b), xMat(n - p, p + 2), res_conf_b, res(5, 3),
               xMatTrend0;

  yVec = x(p, n - 1);
  for (j = 0; j < p; j++)
    assignCol(xMat, j, x(p - 1 - j, n - 2 - j));
  for (i = 0; i < n - p; i++) {
    xMat(i, p) = 1.0;
    xMat(i, p + 1) = i + 1.0;
  }

  if (trend)
    ar = svdfit(xMat, yVec);
  else {
    xMatTrend0 = extractCols(xMat, 0, p);
    ar = svdfit(xMatTrend0, yVec);
    ar = c(ar, 0.0);
  }

  arSum = sum(ar(0, p - 1));
  MatrixDouble arMat(p + 2, 1);
  assignCol(arMat, 0, ar);
  innov = yVec - extractCol(xMat % arMat, 0);

  for (i = 0; i < B; i++) {
    xStar = arTrendBootSim(n, ar[p], ar[p + 1], ar(0, p - 1), innov, b_boot,
                           boot, nStart);
    for (j = 0; j < n_b; j++) {
      res_b = subsampling_ar_Pv(arSum, xStar, p, b_vec[j], tau_pow, stud, trend);
      if (res_b[2] < alpha)
        h(0, j) += 1;
      if (res_b[3] < alpha)
        h(1, j) += 1;
      if (res_b[2] < alpha/2.0)
        h(2, j) += 1;
      if (res_b[3] < alpha/2.0)
        h(3, j) += 1;
      if (res_b[5] < alpha)
        h(4, j) += 1;
    }
  }

  h = h / ((double) B);

  VectorDouble alpha_vec(5);
  alpha_vec[0] = alpha_vec[1] = alpha_vec[4] = alpha;
  alpha_vec[2] = alpha_vec[3] = alpha/2.0;
  for (j = 0; j < 5; j++) {
    diff = fabs(extractRow(h, j) - alpha_vec[j]);
    min_diff = min(diff);
    i = 0;
    while(diff[i] > min_diff + eps) i++;
    b_opt[j] = b_vec[i];
  }

  res(1, 0) = subsampling_ar_conf(x, p, b_opt[0], alpha, tau_pow, stud, trend)(1, 0);
  res(2, 1) = subsampling_ar_conf(x, p, b_opt[1], alpha, tau_pow, stud, trend)(2, 1);
  res(3, 0) = subsampling_ar_conf(x, p, b_opt[2], alpha/2.0, tau_pow, stud, trend)(1, 0);
  res(3, 1) = subsampling_ar_conf(x, p, b_opt[3], alpha/2.0, tau_pow, stud, trend)(2, 1);
  res_conf_b = subsampling_ar_conf(x, p, b_opt[4], alpha/2.0, tau_pow, stud, trend);
  res(0, 0) = res_conf_b(0, 0);
  res(0, 1) = res_conf_b(0, 1);
  res(4, 0) = res_conf_b(4, 0);
  res(4, 1) = res_conf_b(4, 1);

  res(1, 2) = b_opt[0];
  res(2, 2) = b_opt[1];
  res(3, 2) = (b_opt[2] + b_opt[3])/2.0;
  res(4, 2) = b_opt[4];

  return res;
}

MatrixDouble subsampling_ar_conf_b_cal_boot_sim(int n, VectorInt b_vec, double mu,
          double beta, VectorDouble ar, double alpha, int stud, int trend,
          int num_sim,  int B, int b_boot, VectorDouble (*randgen)(int n),
          VectorDouble (*boot) (const VectorDouble &, int, int, int), int n_start)
/*****
  - Simulation Study for AR(p) model with time trend
  - estimate coverage probability and mean and median lengths of normal and
    subsampling intervals with nominal level 1 - alpha
******/
{
  int i, j, p;
  double tau_pow, ar_sum, clt_quantile, alpha_hat, se_n, eps = 0.0000001;
  VectorDouble tau_pow_i(num_sim), x;
  MatrixDouble result(4, 5), length(num_sim, 3), res_i;

  p = ar.getSize();
  ar_sum = sum(ar);

  clt_quantile = my_dinvnr(1 - alpha/2.0);

  //put some general info in first row
  result(0, 0) = (double) n;
  result(0, 1) = ar_sum;
  result(0, 2) = alpha;
  result(0, 3) = (double) num_sim;

	for(i = 0; i < num_sim; i++) {

    cout << i << " ";

    // generate AR(p) sequence
		x = arTrendSim(n, mu, beta, ar, randgen, n_start);

    if (stud)
      tau_pow = 0;
    else {
      //tau_pow = subsampling_ar_range_rate_hat(innov, n, p, trend, 5, 10, 0.01, 0.25, 0.5);
      if (fabs(ar_sum - 1) < eps) tau_pow = 1;
      else tau_pow = 0.5;
      //printf("%f\n", tau_pow);
    }
    tau_pow_i[i] = tau_pow;

    res_i = subsampling_ar_conf_b_cal_boot(x, p, b_vec, alpha, tau_pow, stud,
                        trend, B, b_boot, boot, n_start);
    // CLT
    alpha_hat = res_i(0, 0);
    se_n = res_i(0, 1);
    if (fabs((ar_sum - alpha_hat)/se_n) <= clt_quantile)
      result(1, 4) += 1;
    length(i, 2) = 2 * clt_quantile * se_n;

    // Subsampling intervals
    if (ar_sum > res_i(1, 0))
      result(1, 0) += 1;
    if (ar_sum < res_i(2, 1))
      result(1, 1) += 1;
    for (j = 3; j < 5; j++) {
      if (res_i(j, 0) < ar_sum && ar_sum < res_i(j, 1))
        result(1, j - 1) += 1;
      length(i, j - 3) = res_i(j, 1) - res_i(j, 0);
    }
	}

  for (j = 0; j < 5; j++)
    result(1, j) /= ((double) num_sim);

  for (j = 2; j < 5; j++) {
    result(2, j) = mean(extractCol(length, j - 2));
    result(3, j) = median(extractCol(length, j - 2));
  }

  result(0, 4) = mean(tau_pow_i);

  return(result);
}

MatrixDouble subsampling_ar_fix_b_sim(int n, VectorInt b_vec, double mu,
          double beta, VectorDouble ar, double alpha, int stud, int trend,
          int num_sim, VectorDouble (*randgen)(int n), int n_start)
/*****
  - Simulation Study for AR(p) model with time trend
  - estimate coverage probability and mean and median lengths of normal and
    subsampling intervals with nominal level 1 - alpha
******/
{
  int i, j, l, p, n_b = b_vec.getSize();
  double tau_pow, ar_sum, clt_quantile, alpha_hat, se_n, eps = 0.0000001;
  VectorDouble tau_pow_i(num_sim), x, b_vec_double(b_vec);
  MatrixDouble result(6, n_b), res_i_b;

  p = ar.getSize();
  ar_sum = sum(ar);

  clt_quantile = my_dinvnr(1 - alpha/2.0);

  //put some general info in first row
  result(0, 0) = (double) n;
  result(0, 1) = ar_sum;
  result(0, 2) = alpha;
  result(0, 3) = (double) num_sim;

  assignRow(result, 1, b_vec_double);

	for(i = 0; i < num_sim; i++) {

    cout << i << " ";

    // generate AR(p) sequence
		x = arTrendSim(n, mu, beta, ar, randgen, n_start);

    if (stud)
      tau_pow = 0;
    else {
      //tau_pow = subsampling_ar_range_rate_hat(innov, n, p, trend, 5, 10, 0.01, 0.25, 0.5);
      if (fabs(ar_sum - 1) < eps) tau_pow = 1;
      else tau_pow = 0.5;
      //printf("%f\n", tau_pow);
    }
    tau_pow_i[i] = tau_pow;

    // CLT
    res_i_b = subsampling_ar_conf(x, p, b_vec[1], alpha, tau_pow, stud, trend);
    alpha_hat = res_i_b(0, 0);
    se_n = res_i_b(0, 1);
    if (fabs((ar_sum - alpha_hat)/se_n) <= clt_quantile)
      result(0, 4) += 1;

    // Subsampling intervals
    for (l = 0; l < n_b; l++) {
      res_i_b = subsampling_ar_conf(x, p, b_vec[l], alpha, tau_pow, stud, trend);
      if (ar_sum > res_i_b(1, 0))
        result(2, l) += 1;
      if (ar_sum < res_i_b(2, 1))
        result(3, l) += 1;
      for (j = 3; j < 5; j++) {
        if (res_i_b(j, 0) < ar_sum && ar_sum < res_i_b(j, 1))
          result(j + 1, l) += 1;
      }
   	}
  }
  cout << endl << endl;

  result(0, 4) = result(0, 4) / ((double) num_sim);

  assignSub(result, 2, 0, result(2, 5, 0, n_b - 1) / ((double) num_sim));

  return(result);
}


/*
void stock_conf_sim(int n, double mu, double beta, double *coverage_demean,
                    double *coverage_detrend, int num_sim, int n_start)

  - Simulation Study for AR(1) model with time trend
  - estimate coverage probability of Stocks intervals at level 95%
    for various parameters rho and sample sizes n = 120 and n = 240
  - method depends on computing t-statistic for alpha_AR = 1 and checking
    the tables of Stock

{
  int i, j, k;
  double coverage_rho_dm, coverage_rho_dt, t_stat_dm, *innov, eps = 0.01,
         rho[5] = {1, 0.99, 0.95, 0.9, 0.6}, innov_copy[12], t_stat_dt,
         bounds_dm_120[5][2] = {{-3.12, 0.27}, {-3.18, -0.21},
                                    {-3.51, -1.08}, {-4.00, -1.76}, {0, 0}},
         bounds_dm_240[5][2] = {{-3.12, 0.27}, {-3.24, -0.46},
                                    {-4.00, -1.76}, {-4.88, -2.72}, {0, 0}},
         bounds_dt_120[5][2] = {{-3.73, -0.67}, {-3.75, -0.77},
                                     {-3.92, -1.35}, {-4.31, -1.94}, {0, 0}},
         bounds_dt_240[5][2] = {{-3.73, -0.67}, {-3.78, -0.9},
                                     {-4.31, -1.94}, {-5.8, -2.87}, {0, 0}},
         bounds_dm_500[5][2] = {{-3.12, 0.27}, {-3.43, -0.94}, {-4.96, -2.81},
                                {0, 0}, {0, 0}};

  innov = dvector(0, n + n_start);

  for (j = 0; j < 4; j++) {

    coverage_rho_dt = coverage_rho_dm = 0;

  	for(i = 0; i < num_sim; i++) {
      // gemerate innovations
      rnorm(n + n_start, innov);
      //rnorm_mult(n + n_start, innov, 2);
      //rgarch11(n + n_start, innov, 1, 0.1, 0.8);

      // generate AR(p) sequence and store result in innov[0..n-1]
  		ar_trend_sim(n, mu, beta, &rho[j], 1, innov, n_start);

      for (k = 0; k < 12; k++)
        innov_copy[k] = innov[k];

      // compute demeaned t-stat and check whether it falls within bounds
      subsampling_ar_conf(innov, n, 1, n - 10, 0.05, 0.5, 0, 0);
      t_stat_dm = (innov[0] - 1.0)/innov[1];
      if (fabs(n - 120 + 0.0) < eps)
        if (bounds_dm_120[j][0] < t_stat_dm && t_stat_dm < bounds_dm_120[j][1])
          coverage_rho_dm += 1;
      if (fabs(n - 240 + 0.0) < eps)
        if (bounds_dm_240[j][0] < t_stat_dm && t_stat_dm < bounds_dm_240[j][1])
          coverage_rho_dm += 1;
      if (fabs(n - 500 + 0.0) < eps)
        if (bounds_dm_500[j][0] < t_stat_dm && t_stat_dm < bounds_dm_500[j][1])
          coverage_rho_dm += 1;

      for (k = 0; k < 12; k++)
        innov[k] = innov_copy[k];

      // compute detrended t-stat and check whether it falls within bounds

      subsampling_ar_conf(innov, n, 1, n - 10, 0.05, 0.5, 0, 1);
      t_stat_dt = (innov[0] - 1.0)/innov[1];
      if (fabs(n - 120) < eps)
        if (bounds_dt_120[j][0] < t_stat_dt && t_stat_dt < bounds_dt_120[j][1])
          coverage_rho_dt += 1;
      if (fabs(n - 240) < eps)
        if (bounds_dt_240[j][0] < t_stat_dt && t_stat_dt < bounds_dt_240[j][1])
          coverage_rho_dt += 1;
    }
    coverage_demean[j] = coverage_rho_dm/((double) num_sim);
    coverage_detrend[j] = coverage_rho_dt/((double) num_sim);
  }
  free_dvector(innov, 0, n + n_start);
}
*/
