#include <iostream.h>
#include <iomanip.h>
#include <stdlib.h>
#include <math.h>
//#define NDEBUG
#include <assert.h>

#include "vector.h"
#include "matrix.h"
#include "math.h"

// Default constructor for class VectorInt
VectorInt::VectorInt(int vectorSize)
{
  size = vectorSize;
  ptr = new int[size];
  assert(ptr != 0);

  for (int i = 0; i < size; i++)
    ptr[i] = 0;
}

// Copy constructor for class VectorInt
VectorInt::VectorInt(const VectorInt &init)
{
  size = init.size;
  ptr = new int[size];
  assert(ptr != 0);

  for (int i = 0; i < size; i++)
    ptr[i] = init.ptr[i];
}

// Copy constructor from int * for class VectorInt
VectorInt::VectorInt(const int *vec, const int n)
{
  size = n;
  ptr = new int[size];
  assert(ptr != 0);

  for (int i = 0; i < size; i++)
    ptr[i] = vec[i];
}

// Copy constructor from int * for class VectorInt
VectorInt::VectorInt(const int *vec, const int low, const int up)
{
  size = up - low + 1;
  ptr = new int[size];
  assert(ptr != 0);

  for (int i = 0; i < size; i++)
    ptr[i] = vec[low + i];
}


// Destructor for class VectorInt
VectorInt::~VectorInt()
{
  delete[] ptr;
}

// Get the size of the vector
int VectorInt::getSize() const { return size; }

// Overloaded assignment operator
const VectorInt &VectorInt::operator=(const VectorInt &right)
{
  if (&right != this) {
    delete[] ptr;
    size = right.size;
    ptr = new int[size];
    assert(ptr != 0);

    for (int i = 0; i < size; i++)
      ptr[i] = right.ptr[i];
  }

  return *this;    // enables x = y = z;
}

// Overloaded += operator to add another vector
VectorInt &VectorInt::operator+=(const VectorInt &right)
{
  assert (size == right.getSize());

  for (int i = 0; i < size; i++)
    ptr[i] += right[i];

  return *this;
}

// Overloaded += operator to add scalar;
VectorInt &VectorInt::operator+=(int scalar)
{

  for (int i = 0; i < size; i++)
    ptr[i] += scalar;

  return *this;
}

// Overloaded *= operator to multiply by scalar;
VectorInt &VectorInt::operator*=(int scalar)
{

  for (int i = 0; i < size; i++)
    ptr[i] *= scalar;

  return *this;
}


// Overloaded subscript operator
int &VectorInt::operator[](int subscript) const
{
  //assert(0 <= subscript && subscript < size);

  return ptr[subscript];
}

// return subvector
VectorInt VectorInt::operator()(int low, int high) const
{
  assert(0 <= low && high <= size - 1);
  VectorInt newVector(high - low + 1);

  for (int i = 0; i < high - low + 1; i++)
    newVector[i] = ptr[low + i];

  return newVector;
}

// Overloaded + operator to add integer "element-wise" to VectorInt
VectorInt VectorInt::operator+(int a) const
{
  VectorInt newVector(size);

  for (int i = 0; i < size; i++)
      newVector.ptr[i] = ptr[i] + a;

  return newVector;
}

// Overloaded * operator to multiply VectorInt "element-wise" by integer
VectorInt VectorInt::operator*(int a) const
{
  VectorInt newVector(size);

  for (int i = 0; i < size; i++)
      newVector.ptr[i] = ptr[i] * a;

  return newVector;
}


// Overloaded output operator
ostream &operator<<(ostream &output, const VectorInt &a)
{
  int i;

  for (i = 0; i < a.size; i++) {
    output << a.ptr[i] << ' ';

    if ((i + 1) % 10 == 0)
      output << endl;
  }

  if (i % 10 != 0)
    output << endl;

  return output;    // enables cout << x << y;
}

// Overloaded + operator to add integer "element-wise" to VectorInt
VectorInt operator+(const int a, const VectorInt &vector)
{
  VectorInt newVector(vector);

  for (int i = 0; i < newVector.size; i++)
    newVector.ptr[i] += a;

  return newVector;
}

// Overloaded * operator to multiply VectorInt "element-wise" by integer
VectorInt operator*(const int a, const VectorInt &vector)
{
  VectorInt newVector(vector);

  for (int i = 0; i < newVector.size; i++)
    newVector.ptr[i] *= a;

  return newVector;
}

// Function c concatenates two vectors
VectorInt c(const VectorInt &vec1, const VectorInt &vec2)
{
  int i;
  VectorInt newVector(vec1.getSize() + vec2.getSize());
  for (i = 0; i < vec1.getSize(); i++)
    newVector[i] = vec1[i];
  for (i = 0; i < vec2.getSize(); i++)
    newVector[vec1.getSize() + i] = vec2[i];

  return newVector;
}

// Function c concatenates a vector and a scalar
VectorInt c(const VectorInt &vec1, int scalar)
{
  VectorInt newVector(vec1.getSize() + 1);
  for (int i = 0; i < vec1.getSize(); i++)
    newVector[i] = vec1[i];
  newVector[vec1.getSize()] = scalar;

  return newVector;
}

// Function c concatenates a scalar and a vector
VectorInt c(int scalar, const VectorInt &vec1)
{
  VectorInt newVector(vec1.getSize() + 1);
  newVector[0] = scalar;
  for (int i = 0; i < vec1.getSize(); i++)
    newVector[i + 1] = vec1[i];

  return newVector;
}

// Function c concatenates two scalars
VectorInt c(int scalar1, int scalar2)
{
  VectorInt newVector(2);
  newVector[0] = scalar1;
  newVector[1] = scalar2;

  return newVector;
}

// Function permute returns permuation (in loose sense) of vector
VectorInt permute(const VectorInt &vector, const VectorInt &index)
{
  int i;
  VectorInt vectorPerm(index.getSize());

  for (i = 0; i < index.getSize(); i++)
    vectorPerm[i] = vector[index[i]];

  return vectorPerm;
}

// Function vec returns vector with single element
VectorInt vec(int scalar)
{
  VectorInt newVector(1);

  newVector[0] = scalar;
  return newVector;
}

// Function rep repeats scalar a number of times in vector
VectorInt rep(int scalar, int times)
{
  VectorInt newVector(times);

  for (int i = 0; i < times; i++)
    newVector[i] = scalar;

  return newVector;
}

// Function nullVectorInt returns a vector full of zeroes
VectorInt nullVectorInt(int size)
{
  VectorInt newVector(size);

  return newVector;
}






// Default constructor for class VectorDouble
VectorDouble::VectorDouble(int vectorSize)
{
  size = vectorSize;
  ptr = new double[size];
  assert(ptr != 0);

  for (int i = 0; i < size; i++)
    ptr[i] = 0;
}

// Copy constructor for class VectorDouble
VectorDouble::VectorDouble(const VectorDouble &init)
{
  size = init.size;
  ptr = new double[size];
  assert(ptr != 0);

  for (int i = 0; i < size; i++)
    ptr[i] = init.ptr[i];
}

// Copy constructor for class VectorDouble
VectorDouble::VectorDouble(const VectorInt &init)
{
  size = init.getSize();
  ptr = new double[size];
  assert(ptr != 0);

  for (int i = 0; i < size; i++)
    ptr[i] = (double) init[i];
}

// Copy constructor from double * for class VectorDouble
VectorDouble::VectorDouble(const double *vec, const int n)
{
  size = n;
  ptr = new double[size];
  assert(ptr != 0);

  for (int i = 0; i < size; i++)
    ptr[i] = vec[i];
}

// Copy constructor from double * (partial) for class VectorDouble
VectorDouble::VectorDouble(const double *vec, const int lo, const int hi)
{
  size = hi - lo + 1;
  ptr = new double[size];
  assert(ptr != 0);

  for (int i = 0; i < size; i++)
    ptr[i] = vec[lo + i];
}


// Copy constructor from double * for class VectorDouble
VectorDouble::VectorDouble(const int *vec, const int n)
{
  size = n;
  ptr = new double[size];
  assert(ptr != 0);

  for (int i = 0; i < size; i++)
    ptr[i] = (double) vec[i];
}

// Copy constructor from int * for class VectorDouble
VectorDouble::VectorDouble(const int *vec, const int low, const int up)
{
  size = up - low + 1;
  ptr = new double[size];
  assert(ptr != 0);

  for (int i = 0; i < size; i++)
    ptr[i] = (double) vec[low + i];
}

// Destructor for class VectorDouble
VectorDouble::~VectorDouble()
{
  delete[] ptr;
}

// Get the size of the vector
int VectorDouble::getSize() const { return size; }

// Overloaded assignment operator
const VectorDouble &VectorDouble::operator=(const VectorDouble &right)
{
  if (&right != this) {
    delete[] ptr;
    size = right.size;
    ptr = new double[size];
    assert(ptr != 0);

    for (int i = 0; i < size; i++)
      ptr[i] = right.ptr[i];
  }

  return *this;    // enables x = y = z;
}

// Overloaded assignment operator from VectorInt
const VectorDouble &VectorDouble::operator=(const VectorInt &right)
{
  delete[] ptr;
  size = right.getSize();
  ptr = new double[size];
  assert(ptr != 0);

  for (int i = 0; i < size; i++)
    ptr[i] = (double) right[i];

  return *this;    // enables x = y = z;
}

// Overloaded += operator to add another vector
VectorDouble &VectorDouble::operator+=(const VectorDouble &right)
{
  assert (size == right.getSize());

  for (int i = 0; i < size; i++)
    ptr[i] += right[i];

  return *this;
}

// Overloaded += operator to add scalar;
VectorDouble &VectorDouble::operator+=(double scalar)
{

  for (int i = 0; i < size; i++)
    ptr[i] += scalar;

  return *this;
}

// Overloaded -= operator to subtract scalar;
VectorDouble &VectorDouble::operator-=(double scalar)
{

  for (int i = 0; i < size; i++)
    ptr[i] -= scalar;

  return *this;
}

// Overloaded *= operator to multiply by scalar;
VectorDouble &VectorDouble::operator*=(double scalar)
{

  for (int i = 0; i < size; i++)
    ptr[i] *= scalar;

  return *this;
}

// Overloaded /= operator to divide by scalar;
VectorDouble &VectorDouble::operator/=(double scalar)
{

  for (int i = 0; i < size; i++)
    ptr[i] /= scalar;

  return *this;
}

// Overloaded subscript operator
double &VectorDouble::operator[](int subscript) const
{
  assert(0 <= subscript);
  assert(subscript < size);

  return ptr[subscript];
}

// return subvector
VectorDouble VectorDouble::operator()(int low, int high) const
{
  assert(0 <= low && high <= size - 1);
  VectorDouble newVector(high - low + 1);

  for (int i = 0; i < high - low + 1; i++)
    newVector[i] = ptr[low + i];

  return newVector;
}

// Overloaded + operator to add double "element-wise" to VectorDouble
VectorDouble VectorDouble::operator+(double a) const
{
  VectorDouble newVector(size);

  for (int i = 0; i < size; i++)
      newVector.ptr[i] = ptr[i] + a;

  return newVector;
}

// Overloaded - operator to subtract double "element-wise" from VectorDouble
VectorDouble VectorDouble::operator-(double a) const
{
  VectorDouble newVector(size);

  for (int i = 0; i < size; i++)
      newVector.ptr[i] = ptr[i] - a;

  return newVector;
}

// Overloaded * operator to multiply VectorDouble "element-wise" by double
VectorDouble VectorDouble::operator*(double a) const
{
  VectorDouble newVector(size);

  for (int i = 0; i < size; i++)
      newVector.ptr[i] = ptr[i] * a;

  return newVector;
}

// Overloaded / operator to divide VectorDouble "element-wise" by double
VectorDouble VectorDouble::operator/(double a) const
{
  VectorDouble newVector(size);

  for (int i = 0; i < size; i++)
      newVector.ptr[i] = ptr[i] / a;

  return newVector;
}

// Overloaded ^ operator to raise vector "element-wise" to a power
VectorDouble VectorDouble::operator^(double a) const
{
  VectorDouble newVector(size);

  for (int i = 0; i < size; i++)
      newVector.ptr[i] = pow(ptr[i], a);

  return newVector;
}

// Compute sum of squares
double VectorDouble::sumSquares() const
{
  double sum = 0;

  for (int i = 0; i < size; i++)
    sum += pow(ptr[i], 2);

  return sum;
}


// Overloaded output operator
ostream &operator<<(ostream &output, const VectorDouble &a)
{
  int i;

  cout.flags(ios::showpoint);
  cout.precision(4);

  for (i = 0; i < a.size; i++) {
    cout.width(7);
    output << a.ptr[i] << ' ';

    if ((i + 1) % 8 == 0)
      output << endl;
  }

  if (i % 8 != 0)
    output << endl;

  return output;    // enables cout << x << y;
}

// Overloaded + operator to add double "element-wise" to VectorDouble
VectorDouble operator+(const double a, const VectorDouble &vector)
{
  VectorDouble newVector(vector);

  for (int i = 0; i < newVector.size; i++)
    newVector[i] += a;

  return newVector;
}

// Overloaded * operator to multiply VectorDouble "element-wise" by double
VectorDouble operator*(const double a, const VectorDouble &vector)
{
  VectorDouble newVector(vector);

  for (int i = 0; i < newVector.getSize(); i++)
    newVector[i] *= a;

  return newVector;
}

// Overloaded + operator to add two VectorDoubles "element-wise"
VectorDouble operator+(const VectorDouble &vector1, const VectorDouble &vector2)
{
  assert(vector1.getSize() == vector2.getSize());
  VectorDouble newVector(vector1);

  for (int i = 0; i < newVector.size; i++)
    newVector.ptr[i] += vector2.ptr[i];

  return newVector;
}

// Overloaded - operator to subtract two VectorDoubles "element-wise"
VectorDouble operator-(const VectorDouble &vector1, const VectorDouble &vector2)
{
  assert(vector1.getSize() == vector2.getSize());
  VectorDouble newVector(vector1);

  for (int i = 0; i < newVector.size; i++)
    newVector.ptr[i] -= vector2.ptr[i];

  return newVector;
}

// Overloaded * operator to multiply two VectorDoubles "element-wise"
VectorDouble operator*(const VectorDouble &vector1, const VectorDouble &vector2)
{
  assert(vector1.getSize() == vector2.getSize());
  VectorDouble newVector(vector1);

  for (int i = 0; i < newVector.size; i++)
    newVector.ptr[i] *= vector2.ptr[i];

  return newVector;
}

// Overloaded / operator to divide two VectorDoubles "element-wise"
VectorDouble operator/(const VectorDouble &vector1, const VectorDouble &vector2)
{
  assert(vector1.getSize() == vector2.getSize());
  VectorDouble newVector(vector1);

  for (int i = 0; i < newVector.size; i++)
    newVector.ptr[i] /= vector2.ptr[i];

  return newVector;
}

// Function innerProduct computes inner product between two vectors
double innerProduct(const VectorDouble &vec1, const VectorDouble &vec2)
{
  assert(vec1.getSize() == vec1.getSize());
  int i;
  double inner = 0;

  for (i = 0; i < vec1.getSize(); i++)
    inner += vec1[i] * vec2[i];

  return inner;
}

// Function innerProduct computes inner product between vector and itself
double innerProduct(const VectorDouble &vec)
{
  int i;
  double inner = 0;

  for (i = 0; i < vec.getSize(); i++)
    inner += vec[i] * vec[i];

  return inner;
}

// Function outerProduct computes outer product between two vectors
/*
MatrixDouble outerProduct(const VectorDouble &vec1, const VectorDouble &vec2)
{
  assert(vec1.getSize() == vec1.getSize());
  int i, j, size = vec1.getSize();
  MatrixDouble outer;

  outer = nullMatrixDouble(size, size);
  for (i = 0; i < size; i++)
    for (j = 0; j < size; j++)
      outer(i, j) = vec1[i] * vec2[j];

  return outer;
}  */


// Function c concatenates two vectors
VectorDouble c(const VectorDouble &vec1, const VectorDouble &vec2)
{
  int i;
  VectorDouble newVector(vec1.getSize() + vec2.getSize());
  for (i = 0; i < vec1.getSize(); i++)
    newVector[i] = vec1[i];
  for (i = 0; i < vec2.getSize(); i++)
    newVector[vec1.getSize() + i] = vec2[i];

  return newVector;
}

// Function c concatenates a vector and a scalar
VectorDouble c(const VectorDouble &vec1, double scalar)
{
  VectorDouble newVector(vec1.getSize() + 1);
  for (int i = 0; i < vec1.getSize(); i++)
    newVector[i] = vec1[i];
  newVector[vec1.getSize()] = scalar;

  return newVector;
}

// Function c concatenates a scalar and a vector
VectorDouble c(double scalar, const VectorDouble &vec1)
{
  VectorDouble newVector(vec1.getSize() + 1);
  newVector[0] = scalar;
  for (int i = 0; i < vec1.getSize(); i++)
    newVector[i + 1] = vec1[i];

  return newVector;
}

// Function c concatenates two scalars
VectorDouble c(double scalar1, double scalar2)
{
  VectorDouble newVector(2);

  newVector[0] = scalar1;
  newVector[1] = scalar2;

  return newVector;
}

// Function permute returns permuation (in loose sense) of vector
VectorDouble permute(const VectorDouble &vector, const VectorInt &index)
{
  assert (min(index) >= 0);
  assert (max(index) < vector.getSize());

  int i;
  VectorDouble vectorPerm(index.getSize());

  for (i = 0; i < index.getSize(); i++)
    vectorPerm[i] = vector[index[i]];

  return vectorPerm;
}

// Function rep repeats scalar a number of times in vector
VectorDouble rep(double scalar, int times)
{
  VectorDouble newVector(times);

  for (int i = 0; i < times; i++)
    newVector[i] = scalar;

  return newVector;
}

// Function nullVectorDouble returns a vector full of zeroes
VectorDouble nullVectorDouble(int size)
{
  VectorDouble newVector(size);

  return newVector;
}

// Function vec returns vector with single element
VectorDouble vec(double scalar)
{
  VectorDouble newVector(1);

  newVector[0] = scalar;
  return newVector;
}

// Function assignSub assigns a vector to subvector
void assignSub(VectorDouble &vector, int start, const VectorDouble &other)
{
  assert(0 <= start && start <= vector.getSize() - 1);
  assert(vector.getSize() - start >= other.getSize());

  for (int i = 0; i < other.getSize(); i++)
    vector[start + i] = other[i];
}
